from .levenshtein import levenshtein_distance, levenshtein_distance_async
from .embedding_similarity import (
    create_embedding_similarity_evaluator,
    create_async_embedding_similarity_evaluator,
)

__all__ = [
    "levenshtein_distance",
    "levenshtein_distance_async",
    "create_embedding_similarity_evaluator",
    "create_async_embedding_similarity_evaluator",
]



from openevals.code.base import (
    _create_base_code_evaluator,
    _create_async_base_code_evaluator,
)
from openevals.llm import _create_llm_as_judge_scorer, _create_async_llm_as_judge_scorer
from openevals.prompts import (
    CODE_CORRECTNESS_PROMPT,
    CODE_CORRECTNESS_PROMPT_WITH_REFERENCE_OUTPUTS,
)

from typing import Callable, Optional, Literal, Any

from langchain_core.language_models.chat_models import BaseChatModel

from openevals.types import (
    SimpleEvaluator,
    SimpleAsyncEvaluator,
    RunnableLike,
    ChatCompletionMessage,
    FewShotExample,
)

__all__ = [
    "create_code_llm_as_judge",
    "CODE_CORRECTNESS_PROMPT",
    "CODE_CORRECTNESS_PROMPT_WITH_REFERENCE_OUTPUTS",
]


def create_code_llm_as_judge(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    feedback_key: str = "code_correctness",
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    judge: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
    system: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> SimpleEvaluator:
    """Creates an evaluator that uses an LLM to judge code correctness.

    This function creates an evaluator that extracts code from a response and uses
    an LLM to judge its correctness based on the provided prompt and criteria.

    Takes all the same arguments as the `create_llm_as_judge` function.

    Args:
        prompt: The prompt to use for evaluation. Can be a string, a runnable, or a
            callable that returns a list of chat messages.
        feedback_key: The key to use for storing feedback in the evaluation results.
            Defaults to "code_correctness".
        code_extraction_strategy: Strategy for extracting code from the response.
            Options are "none" (use raw response), "llm" (use LLM to extract code),
            or "markdown_code_blocks" (extract code from markdown blocks).
            Defaults to "none".
        code_extractor: Optional custom function to extract code from the response.
            If provided, overrides the code_extraction_strategy.
        judge: The model client or LangChain chat model to use as the judge.
            If not provided, will use the model specified by the model parameter.
        model: The name of the model to use if judge is not provided.
        system: Optional system message to include in the prompt to the judge.
        continuous: Whether to return a continuous score. If False, returns a
            categorical score based on choices. Defaults to False.
        choices: Optional list of possible score values when continuous is False.
        use_reasoning: Whether to include reasoning in the evaluation output.
            Defaults to True.
        few_shot_examples: Optional list of few-shot examples to include in the prompt.

    Returns:
        A SimpleEvaluator that evaluates code correctness using an LLM as judge.
    """
    scorer = _create_llm_as_judge_scorer(
        prompt=prompt,
        system=system,
        model=model,
        judge=judge,
        continuous=continuous,
        choices=choices,
        use_reasoning=use_reasoning,
        few_shot_examples=few_shot_examples,
    )
    return _create_base_code_evaluator(
        model=model,
        client=judge,
        run_name="code_llm_as_judge",
        feedback_key=feedback_key or "code_correctness",
        scorer=scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )


def create_async_code_llm_as_judge(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    feedback_key: str = "code_correctness",
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    judge: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
    system: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> SimpleAsyncEvaluator:
    """Creates an evaluator that uses an LLM to judge code correctness.

    This function creates an asynchronous evaluator that extracts code from a response and uses
    an LLM to judge its correctness based on the provided prompt and criteria.

    Takes all the same arguments as the `create_async_llm_as_judge` function.

    Args:
        prompt: The prompt to use for evaluation. Can be a string, a runnable, or a
            callable that returns a list of chat messages.
        feedback_key: The key to use for storing feedback in the evaluation results.
            Defaults to "code_correctness".
        code_extraction_strategy: Strategy for extracting code from the response.
            Options are "none" (use raw response), "llm" (use LLM to extract code),
            or "markdown_code_blocks" (extract code from markdown blocks).
            Defaults to "none".
        code_extractor: Optional custom function to extract code from the response.
            If provided, overrides the code_extraction_strategy.
        judge: The model client or LangChain chat model to use as the judge.
            If not provided, will use the model specified by the model parameter.
        model: The name of the model to use if judge is not provided.
        system: Optional system message to include in the prompt to the judge.
        continuous: Whether to return a continuous score. If False, returns a
            categorical score based on choices. Defaults to False.
        choices: Optional list of possible score values when continuous is False.
        use_reasoning: Whether to include reasoning in the evaluation output.
            Defaults to True.
        few_shot_examples: Optional list of few-shot examples to include in the prompt.

    Returns:
        A SimpleAsyncEvaluator that evaluates code correctness using an LLM as judge.
    """
    scorer = _create_async_llm_as_judge_scorer(
        prompt=prompt,
        system=system,
        model=model,
        judge=judge,
        continuous=continuous,
        choices=choices,
        use_reasoning=use_reasoning,
        few_shot_examples=few_shot_examples,
    )
    return _create_async_base_code_evaluator(
        model=model,
        client=judge,
        run_name="code_llm_as_judge",
        feedback_key=feedback_key or "code_correctness",
        scorer=scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )



from openevals.string.levenshtein import (
    levenshtein_distance,
    levenshtein_distance_async,
)
from langsmith import Client
import pytest


@pytest.mark.langsmith
def test_levenshtein():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 2}
    res = levenshtein_distance(outputs=outputs, reference_outputs=reference_outputs)
    assert res["key"] == "levenshtein_distance"
    assert res["score"] == 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
def test_levenshtein_with_different_values():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    res = levenshtein_distance(outputs=outputs, reference_outputs=reference_outputs)
    assert res["key"] == "levenshtein_distance"
    assert res["score"] < 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_levenshtein_async():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 2}
    res = await levenshtein_distance_async(
        outputs=outputs, reference_outputs=reference_outputs
    )
    assert res["key"] == "levenshtein_distance"
    assert res["score"] == 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_levenshtein_with_different_values_async():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    res = await levenshtein_distance_async(
        outputs=outputs, reference_outputs=reference_outputs
    )
    assert res["key"] == "levenshtein_distance"
    assert res["score"] < 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
def test_levenshtein_evaluate():
    client = Client()
    evaluator = levenshtein_distance
    res = client.evaluate(lambda x: x, data="json", evaluators=[evaluator])
    for r in res:
        assert r["evaluation_results"]["results"][0].score is not None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_levenshtein_aevaluate():
    client = Client()

    async def target(inputs):
        return inputs

    evaluator = levenshtein_distance_async
    res = await client.aevaluate(target, data="json", evaluators=[evaluator])
    async for r in res:
        assert r["evaluation_results"]["results"][0].score is not None



from typing import (
    Any,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from typing_extensions import NotRequired, TypedDict


ScoreType = Union[float, bool]


class EvaluatorResult(TypedDict):
    key: str
    score: ScoreType
    comment: Optional[str]
    metadata: Optional[dict]


class SimpleEvaluatorNormal(Protocol):
    def __call__(
        self,
        *,
        inputs: Optional[Any] = None,
        outputs: Any,
        reference_outputs: Optional[Any] = None,
        **kwargs,
    ) -> EvaluatorResult | list[EvaluatorResult]: ...


class SimpleEvaluatorJson(Protocol):
    def __call__(
        self,
        *,
        outputs: Any,
        reference_outputs: Any,
        **kwargs,
    ) -> EvaluatorResult | list[EvaluatorResult]: ...


SimpleEvaluator = Union[SimpleEvaluatorNormal, SimpleEvaluatorJson]


class SimpleAsyncEvaluatorNormal(Protocol):
    async def __call__(
        self,
        *,
        inputs: Optional[Any] = None,
        outputs: Any,
        reference_outputs: Optional[Any] = None,
        **kwargs,
    ) -> EvaluatorResult | list[EvaluatorResult]: ...


class SimpleAsyncEvaluatorJson(Protocol):
    async def __call__(
        self,
        *,
        outputs: Any,
        reference_outputs: Any,
        **kwargs,
    ) -> EvaluatorResult | list[EvaluatorResult]: ...


SimpleAsyncEvaluator = Union[SimpleAsyncEvaluatorNormal, SimpleAsyncEvaluatorJson]


class ChatCompletionMessage(TypedDict):
    content: Union[str, list[dict]]
    role: str
    tool_calls: NotRequired[Optional[list[dict]]]


class ChatCompletion(TypedDict):
    choices: list[dict]


# Few shot example type for evaluator prompts
class FewShotExample(TypedDict):
    inputs: Any
    outputs: Any
    score: float | bool
    reasoning: Optional[str]


@runtime_checkable
class ChatCompletionsClient(Protocol):
    def create(self, **kwargs) -> ChatCompletion: ...


@runtime_checkable
class ModelClient(Protocol):
    @property
    def chat(self) -> type[ChatCompletionsClient]: ...


@runtime_checkable
class RunnableLike(Protocol):
    def invoke(self, inputs: Any, **kwargs) -> Any: ...

    async def ainvoke(self, inputs: Any, **kwargs) -> Any: ...



PYTHON_EVALUATOR_SEPARATOR = "---python-eval-start---"
PYTHON_EVALUATOR_FILE = """
import json
import subprocess
def _parse_pyright_output(stdout: bytes):
    try:
        # Parse the JSON output
        output = json.loads(stdout)

        errors = []
        for error in output.get("generalDiagnostics", []):
            if (error.get("severity", None) == "error"):
                del error["file"]
                errors.append(error)
        score = len(errors) == 0
        return (score, json.dumps(errors))
    except json.JSONDecodeError:
        return (False, f"Failed to parse Pyright output: " + stdout.decode())


def _analyze_with_pyright(
    *,
    pyright_cli_args = None,
):
    result = subprocess.run(
        [
            "pyright",
            "--outputjson",
            "--level",
            "error",  # Only report errors, not warnings
            *(pyright_cli_args or []),
            "outputs.py",
        ],
        capture_output=True,
    )

    print(json.dumps(_parse_pyright_output(result.stdout)))


_analyze_with_pyright()
"""

EXTRACT_IMPORT_NAMES = """
import ast
import sys

BUILTIN_MODULES = set(sys.stdlib_module_names)

def extract_import_names(file_path: str) -> list[str]:
    with open(file_path, "r", encoding="utf-8") as f:
        source_code = f.read()
    
    try:
        tree = ast.parse(source_code)
    except SyntaxError:
        return []

    python_imports = []
    
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for name in node.names:
                import_path = name.name
                if not import_path.startswith((".", "/")):
                    base_package = import_path.split(".")[0]
                    if base_package not in python_imports and base_package not in BUILTIN_MODULES:
                        python_imports.append(base_package)
        elif isinstance(node, ast.ImportFrom):
            if node.module and not node.module.startswith((".", "/")):
                base_package = node.module.split(".")[0]
                if base_package not in python_imports and base_package not in BUILTIN_MODULES:
                    python_imports.append(base_package)

    return python_imports

file_path = "./outputs.py"
imports = extract_import_names(file_path)
print("\\n".join(imports))
"""



CONCISENESS_PROMPT = """You are an expert data labeler evaluating model outputs for conciseness. Your task is to assign a score based on the following rubric:

<Rubric>
  A perfectly concise answer:
  - Contains only the exact information requested.
  - Uses the minimum number of words necessary to convey the complete answer.
  - Omits pleasantries, hedging language, and unnecessary context.
  - Excludes meta-commentary about the answer or the model's capabilities.
  - Avoids redundant information or restatements.
  - Does not include explanations unless explicitly requested.

  When scoring, you should deduct points for:
  - Introductory phrases like "I believe," "I think," or "The answer is."
  - Hedging language like "probably," "likely," or "as far as I know."
  - Unnecessary context or background information.
  - Explanations when not requested.
  - Follow-up questions or offers for more information.
  - Redundant information or restatements.
  - Polite phrases like "hope this helps" or "let me know if you need anything else."
</Rubric>

<Instructions>
  - Carefully read the input and output.
  - Check for any unnecessary elements, particularly those mentioned in the <Rubric> above.
  - The score should reflect how close the response comes to containing only the essential information requested based on the rubric above.
</Instructions>

<Reminder>
  The goal is to reward responses that provide complete answers with absolutely no extraneous information.
</Reminder>

<input>
{inputs}
</input>

<output>
{outputs}
</output>
"""



from openevals.types import EvaluatorResult
from openevals.utils import _run_evaluator, _arun_evaluator

import json
from typing import Any


def _scorer(outputs: Any, reference_outputs: Any) -> bool:
    if outputs is None or reference_outputs is None:
        raise ValueError("Exact match requires both outputs and reference_outputs")
    # Convert both to JSON strings for deep comparison
    outputs_json = json.dumps(outputs, sort_keys=True)
    reference_outputs_json = json.dumps(reference_outputs, sort_keys=True)
    return outputs_json == reference_outputs_json


def exact_match(
    *, outputs: Any, reference_outputs: Any, **kwargs: Any
) -> EvaluatorResult:
    """
    Performs exact matching between input and reference output values.

    Args:
        outputs (Any): Outputs to compare
        reference_outputs (Any): Reference outputs to compare

    Returns:
        EvaluatorResult: Contains match result
    """

    def get_score():
        return _scorer(outputs, reference_outputs)

    res = _run_evaluator(
        run_name="exact_match", scorer=get_score, feedback_key="exact_match"
    )
    if isinstance(res, list):
        return res[0]
    return res


async def exact_match_async(
    *, outputs: Any, reference_outputs: Any, **kwargs: Any
) -> EvaluatorResult:
    """
    Performs exact matching between input and reference output values.

    Args:
        outputs (Any): Outputs to compare
        reference_outputs (Any): Reference outputs to compare

    Returns:
        EvaluatorResult: Contains match result
    """

    async def get_score():
        return _scorer(outputs, reference_outputs)

    res = await _arun_evaluator(
        run_name="exact_match", scorer=get_score, feedback_key="exact_match"
    )
    if isinstance(res, list):
        return res[0]
    return res



from openevals.string.embedding_similarity import (
    create_embedding_similarity_evaluator,
    create_async_embedding_similarity_evaluator,
)
from langsmith import Client
import pytest


@pytest.mark.langsmith
def test_embedding_similarity_documents():
    evaluator = create_embedding_similarity_evaluator()

    inputs = "Where was the first president of FoobarLand born?"

    context = "\n".join(
        [
            "BazQuxLand is a new country located on the dark side of the moon",
            "Space dolphins are native to BazQuxLand",
            "BazQuxLand is a constitutional democracy whose first president was Bagatur Askaryan",
            "The current weather in BazQuxLand is 80 degrees and clear.",
        ]
    )

    result = evaluator(
        outputs=context,
        reference_outputs=inputs,
    )

    print(result)


@pytest.mark.langsmith
def test_embedding_similarity():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 2}
    embedding_similarity = create_embedding_similarity_evaluator()
    res = embedding_similarity(outputs=outputs, reference_outputs=reference_outputs)
    assert res["key"] == "embedding_similarity"
    assert res["score"] == 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
def test_embedding_similarity_with_different_values():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    embedding_similarity = create_embedding_similarity_evaluator()
    res = embedding_similarity(outputs=outputs, reference_outputs=reference_outputs)
    assert res["key"] == "embedding_similarity"
    assert res["score"] < 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_embedding_similarity_async():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 2}
    embedding_similarity = create_async_embedding_similarity_evaluator()
    res = await embedding_similarity(
        outputs=outputs, reference_outputs=reference_outputs
    )
    assert res["key"] == "embedding_similarity"
    assert res["score"] == 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_embedding_similarity_with_different_values_async():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    embedding_similarity = create_async_embedding_similarity_evaluator()
    res = await embedding_similarity(
        outputs=outputs, reference_outputs=reference_outputs
    )
    assert res["key"] == "embedding_similarity"
    assert res["score"] < 1.0
    assert res["comment"] is None


@pytest.mark.langsmith
def test_embedding_similarity_evaluate():
    client = Client()
    evaluator = create_embedding_similarity_evaluator()
    res = client.evaluate(lambda x: x, data="json", evaluators=[evaluator])
    for r in res:
        assert r["evaluation_results"]["results"][0].score is not None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_embedding_similarity_aevaluate():
    client = Client()

    async def target(inputs):
        return inputs

    evaluator = create_async_embedding_similarity_evaluator()
    res = await client.aevaluate(target, data="json", evaluators=[evaluator])
    async for r in res:
        assert r["evaluation_results"]["results"][0].score is not None



import pytest

from langsmith import testing as t

from openevals.llm import create_llm_as_judge
from openevals.prompts.rag_helpfulness import RAG_HELPFULNESS_PROMPT
from openevals.prompts.rag_groundedness import RAG_GROUNDEDNESS_PROMPT
from openevals.prompts.rag_retrieval_relevance import RAG_RETRIEVAL_RELEVANCE_PROMPT


@pytest.mark.langsmith
def test_llm_as_judge_rag_helpfulnesss():
    inputs = {
        "question": "Where was the first president of foobarland born?",
    }
    outputs = {
        "answer": "The first president of foobarland was born in langchainland. His name was Bagatur Askaryan."
    }
    t.log_inputs(inputs)
    t.log_outputs(outputs)
    llm_as_judge = create_llm_as_judge(
        prompt=RAG_HELPFULNESS_PROMPT,
        feedback_key="helpfulness",
        model="openai:gpt-4o-mini",
    )

    eval_result = llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_rag_helpfulness_not_correct():
    inputs = {
        "question": "Where was the first president of foobarland born?",
    }
    outputs = {"answer": "The first president of foobarland was bagatur"}
    t.log_inputs(inputs)
    t.log_outputs(outputs)
    llm_as_judge = create_llm_as_judge(
        prompt=RAG_HELPFULNESS_PROMPT,
        feedback_key="helpfulness",
        model="openai:gpt-4o-mini",
    )

    eval_result = llm_as_judge(inputs=inputs, outputs=outputs)

    assert not eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_rag_groundedness():
    retrieval_evaluator = create_llm_as_judge(
        prompt=RAG_GROUNDEDNESS_PROMPT,
        feedback_key="groundedness",
        model="openai:gpt-4o-mini",
    )

    context = {
        "documents": [
            "FoobarLand is a new country located on the dark side of the moon",
            "Space dolphins are native to FoobarLand",
            "FoobarLand is a constitutional democracy whose first president was Bagatur Askaryan",
            "The current weather in FoobarLand is 80 degrees and clear.",
        ],
    }

    outputs = {
        "answer": "The first president of FoobarLand was Bagatur Askaryan.",
    }

    eval_result = retrieval_evaluator(
        context=context,
        outputs=outputs,
    )

    assert eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_rag_retrieval_relevance():
    retrieval_relevance_evaluator = create_llm_as_judge(
        prompt=RAG_RETRIEVAL_RELEVANCE_PROMPT,
        feedback_key="retrieval_relevance",
        model="openai:gpt-4o-mini",
    )

    inputs = {
        "question": "Where was the first president of FoobarLand born?",
    }

    context = {
        "documents": [
            "FoobarLand is a new country located on the dark side of the moon",
            "Space dolphins are native to FoobarLand",
            "FoobarLand is a constitutional democracy whose first president was Bagatur Askaryan",
            "The current weather in FoobarLand is 80 degrees and clear.",
        ],
    }

    eval_result = retrieval_relevance_evaluator(
        inputs=inputs,
        context=context,
    )

    assert not eval_result["score"]



RAG_RETRIEVAL_RELEVANCE_PROMPT = """You are an expert data labeler evaluating retrieved context for relevance to the input. Your task is to assign a score based on the following rubric:

<Rubric>
- Relevant retrieved context:
  - Contain information that could help answer the input, even if incomplete.
  - May include superfluous information, but it should still be somewhat related to the input.

- Irrelevant retrieved context:
  - Contain no useful information for answering the input.
  - Are entirely unrelated to the input.
  - Contains misleading or incorrect information
  - Contains only tangentially related information with no practical utility
</Rubric>

<Instruction>
- Read and understand the full meaning of the input (including edge cases)
- Formulate a list of facts and relevant context that would be needed to respond to the input
- Analyze the retrieved context to identify:
  - Information directly relevant to answering the query
  - Information partially relevant or contextually helpful
  - Information completely irrelevant to the query
- For each piece of information need identified in the previous step, determine:
  - Whether it is fully addressed by the retrieved documents (cite specific text)
  - Whether it is partially addressed (cite specific text)
  - Whether it is not addressed at all
- Note any facts needed to answer the input that are not found
- Note any context that are completely irrelevant, i.e. contain no relevant facts for answering the input
</Instruction>

<Reminder>  
- Focus solely on whether the retrieved context provides useful information for answering the input.
- Think deeply about why the context is or isn’t relevant.
- Use partial credit where applicable, recognizing context that is somewhat helpful even if incomplete.
</Reminder> 

<inputs>
{inputs}
</inputs>

<retrieved_context>
{context}
</retrieved_context>
"""



HALLUCINATION_PROMPT = """You are an expert data labeler evaluating model outputs for hallucinations. Your task is to assign a score based on the following rubric:

<Rubric>
  A response without hallucinations:
  - Contains only verifiable facts that are directly supported by the input context
  - Makes no unsupported claims or assumptions
  - Does not add speculative or imagined details
  - Maintains perfect accuracy in dates, numbers, and specific details
  - Appropriately indicates uncertainty when information is incomplete
</Rubric>

<Instructions>
  - Read the input context thoroughly
  - Identify all claims made in the output
  - Cross-reference each claim with the input context
  - Note any unsupported or contradictory information
  - Consider the severity and quantity of hallucinations
</Instructions>

<Reminder>
  Focus solely on factual accuracy and support from the input context. Do not consider style, grammar, or presentation in scoring. A shorter, factual response should score higher than a longer response with unsupported claims.
</Reminder>

Use the following context to help you evaluate for hallucinations in the output:

<context>
{context}
</context>

<input>
{inputs}
</input>

<output>
{outputs}
</output>

If available, you may also use the reference outputs below to help you identify hallucinations in the response:

<reference_outputs>
{reference_outputs}
</reference_outputs>
"""



import pytest

from openevals.llm import create_llm_as_judge
from openevals.prompts.conciseness import CONCISENESS_PROMPT


@pytest.mark.langsmith
def test_llm_as_judge_conciseness():
    inputs = {
        "question": "How is the weather in San Francisco?",
    }
    outputs = {"answer": "Sunny and 90 degrees."}
    llm_as_judge = create_llm_as_judge(
        prompt=CONCISENESS_PROMPT,
        feedback_key="conciseness",
        model="openai:o3-mini",
    )
    eval_result = llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_conciseness_not_concise():
    inputs = {
        "question": "How is the weather in San Francisco?",
    }
    outputs = {
        "answer": "Thanks for asking! The current weather in San Francisco is sunny and 90 degrees."
    }
    llm_as_judge = create_llm_as_judge(
        prompt=CONCISENESS_PROMPT,
        feedback_key="conciseness",
        model="openai:o3-mini",
    )
    eval_result = llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]



from typing import Literal, Optional, Dict, Any, Union
from openevals.types import (
    EvaluatorResult,
    SimpleEvaluator,
    SimpleAsyncEvaluator,
)
from openevals.utils import _run_evaluator, _arun_evaluator
from openevals.llm import (
    _create_llm_as_judge_scorer,
    _create_async_llm_as_judge_scorer,
    ModelClient,
)
from langchain_core.language_models.chat_models import BaseChatModel


SYSTEM_PROMPT = """You are an LLM that evaluates the accuracy of structured outputs.
Make sure to evaluate each key the users ask you to evaluate separately. Assign the score
for each key based on its own criteria - DO NOT convolute the scores of different keys.
Also only evaluate the output vs. the reference output based on the criteria. DO NOT EVALUATE
BASED ON ANYTHING ELSE. If the output does not match the reference output in some way that
is not mentioned in the criteria that is not a problem and you should ignore those discrepancies.
Only focus on finding discrepancies based on the criteria. If there is a None value being compared
to a non-None value, you should assign a score of 0.
"""

USER_PROMPT = """Please evaluate the accuracy of the following output keys according to these criteria:
{rubric}
<Outputs>
{outputs}
</Outputs>
<Expected Outputs>
{reference_outputs}
</Expected Outputs>"""


def _prepare_parameters(
    *,
    outputs: Any,
    reference_outputs: Any,
    rubric: Dict[str, str],
    exclude_keys: list[str],
    use_reasoning: bool,
    list_match_mode: Literal[
        "superset", "subset", "same_elements", "ordered"
    ] = "same_elements",
):
    json_schema: dict = {
        "type": "object",
        "title": "structured_match_score",
        "description": "Scores measuring the accuracy of structured outputs",
        "properties": {},
        "required": [],
        "additionalProperties": False,
    }

    scores = {}
    formatted_rubric = ""
    use_list_reducer = False

    if isinstance(outputs, list):
        use_list_reducer = True
        if not isinstance(reference_outputs, list):
            raise ValueError(
                "If outputs is a list, reference_outputs must also be a list"
            )

        # Create mapping dictionaries
        outputs_to_use = {}
        reference_outputs_to_use = {}

        if list_match_mode == "ordered":
            # Outputs/Reference outputs must be in the same order
            for i in range(len(outputs)):
                for key, value in outputs[i].items():
                    outputs_to_use[f"{key}_{i}"] = value
            for i in range(len(reference_outputs)):
                for key, value in reference_outputs[i].items():
                    reference_outputs_to_use[f"{key}_{i}"] = value

        elif list_match_mode == "superset":
            # Match each reference output to the best matching output
            available_outputs = list(range(len(outputs)))
            matched_references = set()

            for i, ref_item in enumerate(reference_outputs):
                best_match_score = -1

                # Try each available output item
                for out_idx in available_outputs:
                    output_item = outputs[out_idx]

                    # Calculate match score based on exact matches of keys
                    match_score = 0
                    for key in ref_item:
                        if (
                            key in output_item
                            and key not in exclude_keys
                            and key not in rubric
                        ):
                            match_score += int(ref_item[key] == output_item[key])

                    # If this is the best match so far, update
                    if match_score > best_match_score:
                        best_match_score = match_score
                        best_match_idx = out_idx

                # If we found a match, use it
                if best_match_idx is not None:
                    for key, value in outputs[best_match_idx].items():
                        outputs_to_use[f"{key}_{i}"] = value

                    for key, value in ref_item.items():
                        reference_outputs_to_use[f"{key}_{i}"] = value

                    # Remove the used output from available options
                    available_outputs.remove(best_match_idx)
                    matched_references.add(i)
                else:
                    # There were extra reference items
                    for key, value in ref_item.items():
                        reference_outputs_to_use[f"{key}_{i}"] = value

        else:  # "same_items" or "subset"
            # Match each output to the best matching reference
            available_references = list(range(len(reference_outputs)))
            matched_outputs = set()

            for i, output_item in enumerate(outputs):
                best_match_idx = None
                best_match_score = -1

                # Try each available reference item
                for ref_idx in available_references:
                    ref_item = reference_outputs[ref_idx]

                    # Calculate match score based on exact matches of keys
                    match_score = 0
                    for key in output_item:
                        if (
                            key in ref_item
                            and key not in exclude_keys
                            and key not in rubric
                        ):
                            match_score += int(output_item[key] == ref_item[key])

                    # If this is the best match so far, update
                    if match_score > best_match_score:
                        best_match_score = match_score
                        best_match_idx = ref_idx

                # If we found a match, use it
                if best_match_idx is not None:
                    for key, value in output_item.items():
                        outputs_to_use[f"{key}_{i}"] = value

                    for key, value in reference_outputs[best_match_idx].items():
                        reference_outputs_to_use[f"{key}_{i}"] = value

                    # Remove the used reference from available options
                    available_references.remove(best_match_idx)
                    matched_outputs.add(i)
                else:
                    # There were extra output items
                    for key, value in output_item.items():
                        outputs_to_use[f"{key}_{i}"] = value

            # For "same_elements" mode: penalize unmatched references
            if list_match_mode == "same_elements":
                for ref_idx in available_references:
                    ref_item = reference_outputs[ref_idx]
                    dummy_idx = len(outputs) + available_references.index(ref_idx)
                    for key, value in ref_item.items():
                        reference_outputs_to_use[f"{key}_{dummy_idx}"] = value

        outputs = outputs_to_use
        reference_outputs = reference_outputs_to_use

    for raw_key, value in outputs.items():
        if use_list_reducer:
            key = raw_key[: raw_key.rfind("_")]
        else:
            key = raw_key
        if key in exclude_keys:
            continue
        if raw_key not in reference_outputs:
            scores[raw_key] = 0
            continue
        if key not in rubric and reference_outputs[raw_key] == value:
            scores[raw_key] = 1
        elif key not in rubric:
            scores[raw_key] = 0
        else:
            key_criteria = rubric[key]
            formatted_rubric += f"Key: {key}, Criteria: {key_criteria}\n"
            if not use_reasoning:
                json_schema["properties"][raw_key] = {
                    "type": "boolean",
                    "description": f"Does the output for key {key}, follow the criteria? {key_criteria}",
                }
            else:
                json_schema["properties"][raw_key] = {
                    "type": "object",
                    "properties": {
                        "reasoning": {
                            "type": "string",
                            "description": f"Reasoning for the score you assigned to key {key}",
                        },
                        "score": {
                            "type": "boolean",
                            "description": f"Does the output for key {key}, follow the criteria? {key_criteria}",
                        },
                    },
                    "required": ["score", "reasoning"],
                    "additionalProperties": False,
                }

    for raw_key, value in reference_outputs.items():
        if use_list_reducer:
            key = raw_key[: raw_key.rfind("_")]
        else:
            key = raw_key
        if key not in exclude_keys and raw_key not in outputs:
            scores[raw_key] = 0

    return (
        outputs,
        reference_outputs,
        json_schema,
        scores,
        formatted_rubric,
        use_list_reducer,
    )


def _aggregate_results(
    *,
    score_key: str,
    scores: dict,
    use_list_reducer: bool,
    aggregator: Optional[Literal["average", "all"]],
    list_aggregator: Literal["average", "all"],
) -> dict:
    if use_list_reducer:
        # First group scores by index
        index_grouped_scores: dict = {}
        for k, v in scores.items():
            index = k[k.rfind("_") + 1 :]
            if index not in index_grouped_scores:
                index_grouped_scores[index] = {}
            base_key = k[: k.rfind("_")]
            index_grouped_scores[index][base_key] = v

        # Apply aggregator to each index group first
        if aggregator == "average":
            index_scores = {}
            for index, group in index_grouped_scores.items():
                if group:  # Skip empty groups
                    total = sum(
                        float(v["score"]) if isinstance(v, dict) else v
                        for v in group.values()
                    )
                    index_scores[index] = total / len(group)
        elif aggregator == "all":
            index_scores = {}
            for index, group in index_grouped_scores.items():
                if group:  # Skip empty groups
                    has_non_one = any(
                        (float(v["score"]) if isinstance(v, dict) else v) != 1
                        for v in group.values()
                    )
                    index_scores[index] = 0 if has_non_one else 1
        else:
            # If no aggregator, keep original structure but grouped by index
            index_scores = index_grouped_scores

        # Then apply list_aggregator across indices
        if list_aggregator == "average":
            if all(isinstance(v, (int, float)) for v in index_scores.values()):
                score = (
                    sum(index_scores.values()) / len(index_scores)
                    if index_scores
                    else 0
                )
                return {f"{score_key}:{aggregator}": score}
            else:
                # For complex structures, do deeper aggregation
                scores_aggregated_across_list: dict = {}
                for _, group in index_scores.items():
                    for key, value in group.items():
                        if key not in scores_aggregated_across_list:
                            scores_aggregated_across_list[key] = []
                        scores_aggregated_across_list[key].append(value)

                # Average across indices for each key
                result = {}
                for key, values in scores_aggregated_across_list.items():
                    if values:
                        result[f"{score_key}:{key}"] = sum(
                            [(v["score"] if isinstance(v, dict) else v) for v in values]
                        ) / len(values)

                return result
        elif list_aggregator == "all":
            if all(isinstance(v, (int, float)) for v in index_scores.values()):
                score = 0 if any(v != 1 for v in index_scores.values()) else 1
                return {f"{score_key}:{aggregator}": score}
            else:
                # For complex structures, do deeper aggregation
                scores_aggregated_across_list = {}
                for _, group in index_scores.items():
                    for key, value in group.items():
                        if key not in scores_aggregated_across_list:
                            scores_aggregated_across_list[key] = []
                        scores_aggregated_across_list[key].append(value)

                # Apply 'all' aggregation across indices for each key
                result = {}
                for key, values in scores_aggregated_across_list.items():
                    result[f"{score_key}:{key}"] = (
                        0 if any(v != 1 for v in values) else 1
                    )

                return result

    # Handle non-list case or fallback
    if aggregator == "average":
        score = (
            sum(
                float(v["score"]) if isinstance(v, dict) else v for v in scores.values()
            )
            / len(scores)
            if scores
            else 0
        )
        return {f"{score_key}:{aggregator}": score}
    elif aggregator == "all":
        score = (
            0
            if any(
                (float(v["score"]) if isinstance(v, dict) else v) != 1
                for v in scores.values()
            )
            else 1
        )
        return {f"{score_key}:{aggregator}": score}
    else:
        # No aggregator, return scores as-is
        results = {}
        for key, value in scores.items():
            results[f"{score_key}:{key}"] = value
        return results


def create_json_match_evaluator(
    *,
    aggregator: Optional[Literal["average", "all"]] = None,
    list_aggregator: Literal["average", "all"] = "all",
    rubric: Dict[str, str] = {},
    exclude_keys: list[str] = [],
    judge: Optional[
        Union[
            ModelClient,
            BaseChatModel,
        ]
    ] = None,
    model: Optional[str] = None,
    use_reasoning: bool = True,
    list_match_mode: Literal[
        "superset", "subset", "same_elements", "ordered"
    ] = "same_elements",
) -> SimpleEvaluator:
    """
    Create an evaluator to evaluate the accuracy of structured outputs.

    Parameters:
        aggregator (Optional[Literal["average", "all"]]): The aggregation method to use for combining the keys of each structured object.
            Defaults to None. If None, will return a single EvaluatorResult for each key that appears in either
            the outputs or the reference_outputs or both. If "average", will return a single EvaluatorResult that
            is the average of the feedback for each key in the outputs/reference_outputs. If "all", will return
            a single EvaluatorResult that is a combined and statement of the feedback for each key in the outputs/reference_outputs.
            If "all"/"average" the feedback key returned will be called "json_match"
        list_aggregator (Literal["average", "all"]): The aggregation method to use when evaluating a list of outputs.
            Defaults to "all". If "all", the score for a single feedback key will be a combined and statement of the scores for
            that key across all elements of the list. If "average", the score for a single feedback key will be the
            average of the scores for that key across all elements of the list
        rubric (Optional[Dict[str,str]]): The rubric to use for the judge. Each entry of the dict is a
            key/value pair where the key is the structured output key and the value is the criteria for the LLM to
            evaluate that key on against the reference output.
        exclude_keys (Optional[list[str]]): The keys to exclude from the evaluation. Use this if there are
            keys in your structured output you don't care about evaluating. Every key not in `exclude_keys` or in `rubric`
            will be evaluated for exact match with the reference output.
        judge (ModelClient or BaseChatModel): The judge to use for the evaluation.
        model (str): The model to use for the evaluation.
        use_reasoning (bool): Whether to use reasoning for the keys in `rubric`. Defaults to True.
        match_mode (Literal["subset_of_output", "subset_of_reference", "exact"]): The mode to use for the evaluation.
            Defaults to "exact". If "exact", the evaluation will match every element of the outputs with a corresponding
            element of the reference outputs and vice versa. If "subset_of_reference", the evaluation will match every element
            of the outputs with a corresponding element of the reference outputs. If "subset_of_output", the evaluation will match
            every element of the reference outputs with a corresponding element of the outputs.

    Returns:
        A function that takes in outputs and reference_outputs and returns an EvaluatorResult or list of EvaluatorResults.
    """
    if not judge and not model and len(rubric) != 0:
        raise ValueError("When passing rubric, either judge or model must be provided")
    if len(rubric) == 0 and (judge or model):
        raise ValueError(
            "When not passing rubric, either judge or model must be provided"
        )

    def wrapped_evaluator(
        *,
        outputs: Any,
        reference_outputs: Any,
        **kwargs,
    ) -> list[EvaluatorResult]:
        def _scorer(
            *,
            outputs: Any,
            reference_outputs: Any,
            rubric: Dict[str, str] = {},
            exclude_keys: list[str] = [],
            use_reasoning: bool = True,
        ) -> dict | float | tuple:
            (
                outputs,
                reference_outputs,
                json_schema,
                scores,
                formatted_rubric,
                use_list_reducer,
            ) = _prepare_parameters(
                outputs=outputs,
                reference_outputs=reference_outputs,
                rubric=rubric,
                exclude_keys=exclude_keys,
                use_reasoning=use_reasoning,
                list_match_mode=list_match_mode,
            )

            scorer = None
            if len(formatted_rubric) > 0:
                output_keys = "\n".join(
                    [f"{key}: {outputs[key]}" for key in json_schema["properties"]]
                )
                expected_output_keys = "\n".join(
                    [
                        f"{key}: {reference_outputs[key]}"
                        for key in json_schema["properties"]
                    ]
                )
                scorer = _create_llm_as_judge_scorer(
                    system=SYSTEM_PROMPT,
                    prompt=USER_PROMPT,
                    schema=json_schema,
                    judge=judge,
                    model=model,
                )
            else:
                formatted_rubric, output_keys, expected_output_keys = None, None, None
            if scorer is not None:
                llm_scores = scorer(
                    outputs=output_keys,
                    reference_outputs=expected_output_keys,
                    rubric=rubric,
                )
                scores.update(llm_scores)

            return _aggregate_results(
                score_key="json_match",
                scores=scores,
                use_list_reducer=use_list_reducer,
                aggregator=aggregator,
                list_aggregator=list_aggregator,
            )

        return _run_evaluator(
            run_name="json_match_evaluator",
            scorer=_scorer,
            feedback_key="json_match",
            rubric=rubric,
            outputs=outputs,
            reference_outputs=reference_outputs,
            exclude_keys=exclude_keys,
            use_reasoning=use_reasoning,
            **kwargs,
        )  # type: ignore

    return wrapped_evaluator


def create_async_json_match_evaluator(
    *,
    aggregator: Optional[Literal["average", "all"]] = None,
    list_aggregator: Literal["average", "all"] = "all",
    rubric: Dict[str, str] = {},
    exclude_keys: list[str] = [],
    judge: Optional[
        Union[
            ModelClient,
            BaseChatModel,
        ]
    ] = None,
    model: Optional[str] = None,
    use_reasoning: bool = True,
    list_match_mode: Literal[
        "superset", "subset", "same_elements", "ordered"
    ] = "same_elements",
) -> SimpleAsyncEvaluator:
    """
    Create an evaluator to evaluate the accuracy of structured outputs.

    Parameters:
        aggregator (Optional[Literal["average", "all"]]): The aggregation method to use for combining the keys of each structured object.
            Defaults to None. If None, will return a single EvaluatorResult for each key that appears in either
            the outputs or the reference_outputs or both. If "average", will return a single EvaluatorResult that
            is the average of the feedback for each key in the outputs/reference_outputs. If "all", will return
            a single EvaluatorResult that is a combined and statement of the feedback for each key in the outputs/reference_outputs.
            If "all"/"average" the feedback key returned will be called "structured_match_score
        list_aggregator (Literal["average", "all"]): The aggregation method to use when evaluating a list of outputs.
            Defaults to "all". If "all", the score for a single feedback key will be a combined and statement of the scores for
            that key across all elements of the list. If "average", the score for a single feedback key will be the
            average of the scores for that key across all elements of the list
        rubric (Optional[Dict[str,str]]): The rubric to use for the judge. Each entry of the dict is a
            key/value pair where the key is the structured output key and the value is the criteria for the LLM to
            evaluate that key on against the reference output.
        exclude_keys (Optional[list[str]]): The keys to exclude from the evaluation. Use this if there are
            keys in your structured output you don't care about evaluating. Every key not in `exclude_keys` or in `rubric`
            will be evaluated for exact match with the reference output.
        judge (ModelClient or BaseChatModel): The judge to use for the evaluation.
        model (str): The model to use for the evaluation.
        use_reasoning (bool): Whether to use reasoning for the keys in `rubric`. Defaults to True.
        match_mode (Literal["subset_of_output", "subset_of_reference", "exact"]): The mode to use for the evaluation.
            Defaults to "exact". If "exact", the evaluation will match every element of the outputs with a corresponding
            element of the reference outputs and vice versa. If "subset_of_reference", the evaluation will match every element
            of the outputs with a corresponding element of the reference outputs. If "subset_of_output", the evaluation will match
            every element of the reference outputs with a corresponding element of the outputs.

    Returns:
        A function that takes in outputs and reference_outputs and returns an EvaluatorResult or list of EvaluatorResults.
    """
    if not judge and not model and len(rubric) != 0:
        raise ValueError("When passing rubric, either judge or model must be provided")
    if len(rubric) == 0 and (judge or model):
        raise ValueError(
            "When not passing rubric, either judge or model must be provided"
        )

    async def wrapped_evaluator(
        *,
        outputs: Any,
        reference_outputs: Any,
        **kwargs,
    ) -> EvaluatorResult | list[EvaluatorResult]:
        async def _ascorer(
            *,
            outputs: Any,
            reference_outputs: Any,
            rubric: Dict[str, str] = {},
            exclude_keys: list[str] = [],
            use_reasoning: bool = True,
        ) -> dict | float | tuple:
            (
                outputs,
                reference_outputs,
                json_schema,
                scores,
                formatted_rubric,
                use_list_reducer,
            ) = _prepare_parameters(
                outputs=outputs,
                reference_outputs=reference_outputs,
                rubric=rubric,
                exclude_keys=exclude_keys,
                use_reasoning=use_reasoning,
                list_match_mode=list_match_mode,
            )

            scorer = None
            if len(formatted_rubric) > 0:
                output_keys = "\n".join(
                    [f"{key}: {outputs[key]}" for key in json_schema["properties"]]
                )
                expected_output_keys = "\n".join(
                    [
                        f"{key}: {reference_outputs[key]}"
                        for key in json_schema["properties"]
                    ]
                )
                scorer = _create_async_llm_as_judge_scorer(
                    system=SYSTEM_PROMPT,
                    prompt=USER_PROMPT,
                    schema=json_schema,
                    judge=judge,
                    model=model,
                )
            else:
                formatted_rubric, output_keys, expected_output_keys = None, None, None
            if scorer is not None:
                scorer_res = scorer(
                    outputs=output_keys,
                    reference_outputs=expected_output_keys,
                    rubric=rubric,
                )
                if hasattr(scorer_res, "__await__"):
                    llm_scores = await scorer_res
                else:
                    llm_scores = scorer_res
                scores.update(llm_scores)

            return _aggregate_results(
                score_key="json_match",
                scores=scores,
                use_list_reducer=use_list_reducer,
                aggregator=aggregator,
                list_aggregator=list_aggregator,
            )

        return await _arun_evaluator(
            run_name="structured_match_evaluator",
            scorer=_ascorer,
            feedback_key="structured_match_score",
            rubric=rubric,
            outputs=outputs,
            reference_outputs=reference_outputs,
            exclude_keys=exclude_keys,
            use_reasoning=use_reasoning,
            **kwargs,
        )

    return wrapped_evaluator



RAG_HELPFULNESS_PROMPT = """You are an expert evaluator assessing how helpful and relevant outputs are in addressing an input query. Your evaluation should focus on the following criteria:

<Rubric>
A helpful and relevant output should:
- Directly address the core question or need in the input
- Provide accurate and necessary information
- Be appropriately detailed for the query's scope
- May contradict your built-in knowledge but still be correct for the given context

An unhelpful or irrelevant output:
- Fails to address the main question
- Contains primarily unrelated information
</Rubric>

<Instruction>
- Read and understand the full meaning of the input (including edge cases)
- Identify any implicit requirements or context
- Identify the expected scope of the answer

- Analyze the output to identify:
  - How well it addresses the core question
  - The relevance of included information
  - Any critical missing information
  - Any extraneous or unhelpful content
</Instruction>

<Reminder>
- Evaluate based on practical usefulness to the query
- Consider both direct relevance and helpful context
- Identify specific strengths and weaknesses in the response
- Provide clear reasoning for your assessment
- Remember that correct information may differ from your built-in knowledge based on internal retrieved context
</Reminder> 

<inputs>
{inputs}
</inputs>

<outputs>
{outputs}
</outputs>
"""



import json
from openevals.types import (
    EvaluatorResult,
    SimpleEvaluator,
    SimpleAsyncEvaluator,
)
from openevals.utils import _run_evaluator, _arun_evaluator
from langchain.embeddings import init_embeddings

from typing import Any


def _handle_embedding_outputs(
    algorithm: str, received_embedding: list[float], expected_embedding: list[float]
) -> float:
    def dot_product(v1, v2):
        return sum(a * b for a, b in zip(v1, v2))

    def vector_magnitude(v):
        return (sum(x * x for x in v)) ** 0.5

    def cosine_similarity(v1, v2):
        dot_prod = dot_product(v1, v2)
        magnitude1 = vector_magnitude(v1)
        magnitude2 = vector_magnitude(v2)
        return dot_prod / (magnitude1 * magnitude2)

    # Calculate similarity based on chosen algorithm
    if algorithm == "cosine":
        similarity = cosine_similarity(received_embedding, expected_embedding)
    else:
        similarity = dot_product(received_embedding, expected_embedding)
    return round(similarity, 2)


def create_embedding_similarity_evaluator(
    *, model: str = "openai:text-embedding-3-small", algorithm: str = "cosine"
) -> SimpleEvaluator:
    """
    Create an evaluator that compares outputs and reference outputs for similarity by text embedding distance.

    Args:
        model (str): The model to use for embedding similarity
        algorithm (str): The algorithm to use for embedding similarity. Defaults to 'cosine'. 'dot_product' is also supported.

    Returns:
        EvaluatorResult: A score representing the embedding similarity
    """

    if algorithm != "cosine" and algorithm != "dot_product":
        raise ValueError(
            f"Unsupported algorithm: {algorithm}. Only 'cosine' and 'dot_product' are supported."
        )

    def wrapped_evaluator(
        *, outputs: Any, reference_outputs: Any, **kwargs: Any
    ) -> EvaluatorResult:
        if outputs is None or reference_outputs is None:
            raise ValueError(
                "Embedding similarity requires both outputs and reference_outputs"
            )

        if not isinstance(outputs, str):
            outputs = json.dumps(outputs)
        if not isinstance(reference_outputs, str):
            reference_outputs = json.dumps(reference_outputs)

        def get_score():
            embeddings = init_embeddings(model)
            received_embedding = embeddings.embed_query(outputs)
            expected_embedding = embeddings.embed_query(reference_outputs)

            similarity = _handle_embedding_outputs(
                algorithm, received_embedding, expected_embedding
            )

            return similarity

        res = _run_evaluator(
            run_name="embedding_similarity",
            scorer=get_score,
            feedback_key="embedding_similarity",
        )
        if isinstance(res, list):
            return res[0]
        return res

    return wrapped_evaluator  # type: ignore


def create_async_embedding_similarity_evaluator(
    *, model: str = "openai:text-embedding-3-small", algorithm: str = "cosine"
) -> SimpleAsyncEvaluator:
    """
    Create an evaluator that compares the actual output and reference output for similarity by text embedding distance.

    Args:
        model (str): The model to use for embedding similarity
        algorithm (str): The algorithm to use for embedding similarity. Defaults to 'cosine'. 'dot_product' is also supported.

    Returns:
        EvaluatorResult: A score representing the embedding similarity
    """
    if algorithm != "cosine" and algorithm != "dot_product":
        raise ValueError(
            f"Unsupported algorithm: {algorithm}. Only 'cosine' and 'dot_product' are supported."
        )

    async def wrapped_evaluator(
        *, outputs: Any, reference_outputs: Any, **kwargs: Any
    ) -> EvaluatorResult:
        if outputs is None or reference_outputs is None:
            raise ValueError(
                "Embedding similarity requires both outputs and reference_outputs"
            )

        if not isinstance(outputs, str):
            outputs = json.dumps(outputs)
        if not isinstance(reference_outputs, str):
            reference_outputs = json.dumps(reference_outputs)

        async def get_score():
            from langchain.embeddings import init_embeddings

            embeddings = init_embeddings(model)
            received_embedding = await embeddings.aembed_query(outputs)
            expected_embedding = await embeddings.aembed_query(reference_outputs)

            similarity = _handle_embedding_outputs(
                algorithm, received_embedding, expected_embedding
            )

            return similarity

        res = await _arun_evaluator(
            run_name="embedding_similarity",
            scorer=get_score,
            feedback_key="embedding_similarity",
        )
        if isinstance(res, list):
            return res[0]
        return res

    return wrapped_evaluator  # type: ignore



from __future__ import annotations
import asyncio

from langsmith import testing as t, get_current_run_tree, traceable
from langsmith.testing._internal import _TEST_CASE
from typing import Any, Callable, TYPE_CHECKING, Union, Optional

from openevals.types import ChatCompletionMessage, EvaluatorResult

from langchain_core.messages.utils import convert_to_openai_messages
from langchain_core.messages import BaseMessage

__all__ = [
    "_chat_completion_messages_to_string",
    "_run_evaluator",
    "_arun_evaluator",
    "_normalize_to_openai_messages_list",
    "_normalize_final_app_outputs_as_string",
]


if TYPE_CHECKING:
    from langchain_core.messages import BaseMessage


def _convert_to_openai_message(
    message: ChatCompletionMessage | BaseMessage | dict,
) -> ChatCompletionMessage:
    if isinstance(message, dict):
        return message  # type: ignore
    else:
        if not isinstance(message, BaseMessage):
            raise ValueError(f"Expected BaseMessage, got {type(message)}")
        return convert_to_openai_messages([message])[0]  # type: ignore


def _normalize_to_openai_messages_list(
    messages: Optional[Union[list[ChatCompletionMessage], list[BaseMessage], dict]],
) -> list[ChatCompletionMessage]:
    if messages is None:
        return []
    if isinstance(messages, dict):
        if "messages" in messages:
            messages = messages["messages"]
        else:
            raise ValueError("if messages is a dict, it must contain a 'messages' key")
    return [_convert_to_openai_message(message) for message in messages]  # type: ignore


# Helper function to process individual scores
def _process_score(key: str, value: Any) -> tuple[float, str | None, dict | None]:
    if isinstance(value, dict):
        if "score" in value:
            return value["score"], value.get("reasoning"), value.get("metadata", None)  # type: ignore
        raise ValueError(
            f"Expected a dictionary with keys 'score' and 'reasoning', but got {value}"
        )
    return value, None, None


def _add_metadata_to_run_tree(
    run_name: str,
    framework: str | None = None,
    results: Optional[Union[dict, list[dict]]] = None,
):
    rt = get_current_run_tree()
    if rt is not None:
        if results is not None:
            if isinstance(results, list):
                for result in results:
                    if result.get("metadata", None) is not None:
                        rt.metadata.update(result.get("metadata", None))
            else:
                if results.get("metadata", None) is not None:
                    rt.metadata.update(results.get("metadata", None))
        rt.metadata["__ls_framework"] = framework
        rt.metadata["__ls_evaluator"] = run_name
        rt.metadata["__ls_language"] = "python"


def _run_evaluator(
    *,
    run_name: str,
    scorer: Callable,
    feedback_key: str,
    ls_framework: str = "openevals",
    **kwargs: Any,
) -> EvaluatorResult | list[EvaluatorResult]:
    @traceable(name=run_name)
    def _run_scorer(**kwargs: Any):
        # Get the initial score
        score = scorer(**kwargs)

        # Collect all results first
        if isinstance(score, dict):
            results = []
            # Handle dictionary of scores
            for key, value in score.items():
                key_score, reasoning, metadata = _process_score(key, value)
                results.append(
                    EvaluatorResult(
                        key=key, score=key_score, comment=reasoning, metadata=metadata
                    )
                )
            return results
        else:
            # Handle single score
            if isinstance(score, tuple):
                if len(score) == 3:
                    score, reasoning, metadata = score
                elif len(score) == 2:
                    score, reasoning = score
                    metadata = None
                else:
                    raise ValueError(f"Expected a tuple of length 2 or 3, got {score}")
            else:
                reasoning = None
                metadata = None
            return EvaluatorResult(
                key=feedback_key, score=score, comment=reasoning, metadata=metadata
            )

    # Log feedback if in test case
    if _TEST_CASE.get():
        with t.trace_feedback(name=run_name):
            results = _run_scorer(**kwargs)
            _add_metadata_to_run_tree(run_name, ls_framework, results)
            if isinstance(results, list):
                for result in results:
                    t.log_feedback(
                        key=result["key"],
                        score=result["score"],
                        comment=result["comment"],
                    )
            else:
                _add_metadata_to_run_tree(run_name, ls_framework, results)
                t.log_feedback(
                    key=results["key"],
                    score=results["score"],
                    comment=results["comment"],
                )
    else:
        results = _run_scorer(**kwargs)
        _add_metadata_to_run_tree(run_name, ls_framework, results)

    # Return single result or list of results
    return results


async def _arun_evaluator(
    *,
    run_name: str,
    scorer: Callable,
    feedback_key: str,
    ls_framework: str = "openevals",
    **kwargs: Any,
) -> EvaluatorResult | list[EvaluatorResult]:
    @traceable(name=run_name)
    async def _arun_scorer(**kwargs: Any):
        # Get the initial score
        if asyncio.iscoroutinefunction(scorer):
            score = await scorer(**kwargs)
        else:
            score = scorer(**kwargs)

        # Collect all results first
        if isinstance(score, dict):
            results = []
            # Handle dictionary of scores
            for key, value in score.items():
                key_score, reasoning, metadata = _process_score(key, value)
                results.append(
                    EvaluatorResult(
                        key=key, score=key_score, comment=reasoning, metadata=metadata
                    )
                )
            return results
        else:
            # Handle single score
            if isinstance(score, tuple):
                if len(score) == 3:
                    score, reasoning, metadata = score
                elif len(score) == 2:
                    score, reasoning = score
                    metadata = None
                else:
                    raise ValueError(f"Expected a tuple of length 2 or 3, got {score}")
            else:
                reasoning = None
                metadata = None
            return EvaluatorResult(
                key=feedback_key, score=score, comment=reasoning, metadata=metadata
            )

    # Log feedback if in test case
    if _TEST_CASE.get():
        with t.trace_feedback(name=run_name):
            results = await _arun_scorer(**kwargs)
            _add_metadata_to_run_tree(run_name, ls_framework, results)
            if isinstance(results, list):
                for result in results:
                    t.log_feedback(
                        key=result["key"],
                        score=result["score"],
                        comment=result["comment"],
                    )
            else:
                _add_metadata_to_run_tree(run_name, ls_framework, results)
                t.log_feedback(
                    key=results["key"],
                    score=results["score"],
                    comment=results["comment"],
                )
    else:
        results = await _arun_scorer(**kwargs)
        _add_metadata_to_run_tree(run_name, ls_framework, results)

    # Return single result or list of results
    return results


def _chat_completion_messages_to_string(messages: list[ChatCompletionMessage]) -> str:
    def format_message(message: ChatCompletionMessage) -> str:
        content = message.get("content", "")  # Handle None content

        # Handle tool/function calls
        tool_calls = message.get("tool_calls") or []
        if message.get("tool_calls", None):
            tool_calls_str = "\n".join(
                f"<tool_call>\n"
                f"<name>{call.get('function', {}).get('name', '')}</name>\n"
                f"<arguments>{call.get('function', {}).get('arguments', '')}</arguments>\n"
                f"</tool_call>"
                for call in tool_calls
            )
            content = f"{content}\n{tool_calls_str}" if content else tool_calls_str

        # Handle tool call results
        if message.get("tool_call_id", None):
            content = (
                f"<tool_result>\n"
                f"<id>{message.get('tool_call_id')}</id>\n"
                f"<content>{content}</content>\n"
                f"</tool_result>"
            )

        return f"<{message.get('role', '')}>\n{content}\n</{message.get('role', '')}>"

    return "\n\n".join(format_message(message) for message in messages)


def _normalize_final_app_outputs_as_string(outputs: Union[str, dict]):
    if isinstance(outputs, str):
        return outputs
    elif isinstance(outputs, dict):
        if "content" in outputs:
            converted_message = _convert_to_openai_message(outputs)
            return converted_message["content"]
        elif "messages" in outputs and isinstance(outputs["messages"], list):
            final_message = _convert_to_openai_message(outputs["messages"][-1])
            return final_message["content"]
        else:
            raise ValueError(
                f"Expected a string, dictionary with a 'content' key or a 'messages' key with a list of messages, but got {outputs}"
            )
    else:
        raise ValueError(f"Expected str or dict, got {type(outputs)}")



from .match import (
    create_json_match_evaluator,
    create_async_json_match_evaluator,
    SYSTEM_PROMPT,
    USER_PROMPT,
)

__all__ = [
    "create_json_match_evaluator",
    "create_async_json_match_evaluator",
    "SYSTEM_PROMPT",
    "USER_PROMPT",
]



RAG_GROUNDEDNESS_PROMPT = """You are an expert data labeler assessing how well LLM output aligns with and is supported by the retrieved context. Your evaluation should focus on the following criteria:

<Rubric>
A well-grounded output should:
- Make claims that are directly supported by the retrieved context
- Stay within the scope of information provided in the context
- Maintain the same meaning and intent as the source material
- Not introduce external facts or unsupported assertions outside of basic facts (2 + 2 = 4)

An ungrounded output:
- Makes claims without support from the context
- Contradicts the retrieved information
- Includes speculation or external knowledge outside of basic facts
- Distorts or misrepresents the context
</Rubric>

<Instruction>
- Compare the output against the retrieved context carefully
- Identify claims, statements, and assertions in the output
- For each claim, locate supporting evidence in the context
- Check for:
  - Direct statements from context
  - Valid inferences from context
  - Unsupported additions
  - Contradictions with context

- Note any instances where the output:
  - Extends beyond the context
  - Combines information incorrectly
  - Makes logical leaps
</Instruction>

<Reminder>
- Focus solely on alignment with provided context
- Ignore whether external knowledge suggests different facts
- Consider both explicit and implicit claims
- Provide specific examples of grounded/ungrounded content
- Remember that correct grounding means staying true to the context, even if context conflicts with common knowledge
</Reminder>

<context>
{context}
</context>

<outputs>
{outputs}
</outputs>
"""



import json
from openevals.types import EvaluatorResult
from openevals.utils import _run_evaluator, _arun_evaluator
from typing import Any


def _scorer(outputs: Any, reference_outputs: Any) -> float:
    if outputs is None or reference_outputs is None:
        raise ValueError(
            "Levenshtein distance requires both outputs and reference_outputs"
        )
    if not isinstance(outputs, str):
        outputs = json.dumps(outputs)
    if not isinstance(reference_outputs, str):
        reference_outputs = json.dumps(reference_outputs)
    # Create a matrix of size (m+1)x(n+1) where m and n are the string lengths
    m, n = len(outputs), len(reference_outputs)
    dp = [[0] * (n + 1) for _ in range(m + 1)]

    # Initialize first row and column
    for i in range(m + 1):
        dp[i][0] = i
    for j in range(n + 1):
        dp[0][j] = j

    # Fill the matrix
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if outputs[i - 1] == reference_outputs[j - 1]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = min(
                    dp[i - 1][j] + 1,  # deletion
                    dp[i][j - 1] + 1,  # insertion
                    dp[i - 1][j - 1] + 1,  # substitution
                )

    # Calculate the distance and normalize it to a score between 0 and 1
    distance = dp[m][n]
    max_length = max(m, n)
    score = 1.0 - (distance / max_length) if max_length > 0 else 1.0
    return score


def levenshtein_distance(
    *, outputs: Any, reference_outputs: Any, **kwargs: Any
) -> EvaluatorResult:
    """
    Evaluates the actual output and reference output for similarity by Levenshtein distance.

    Args:
        outputs (Any): Outputs to compare
        reference_outputs (Any): Reference outputs to compare

    Returns:
        EvaluatorResult: Contains match result with score between 0.0 and 1.0, where 1.0 indicates
        an exact match and lower values indicate greater differences
    """

    def get_score():
        return _scorer(outputs, reference_outputs)

    res = _run_evaluator(
        run_name="levenshtein_distance",
        scorer=get_score,
        feedback_key="levenshtein_distance",
    )
    if isinstance(res, list):
        return res[0]
    return res


async def levenshtein_distance_async(
    *, outputs: Any, reference_outputs: Any, **kwargs: Any
) -> EvaluatorResult:
    """
    Evaluates the actual output and reference output for similarity by Levenshtein distance.

    Args:
        outputs (Any): Outputs to compare
        reference_outputs (Any): Reference outputs to compare

    Returns:
        EvaluatorResult: Contains match result with score between 0.0 and 1.0, where 1.0 indicates
        an exact match and lower values indicate greater differences
    """

    async def get_score():
        return _scorer(outputs, reference_outputs)

    res = await _arun_evaluator(
        run_name="levenshtein_distance",
        scorer=get_score,
        feedback_key="levenshtein_distance",
    )
    if isinstance(res, list):
        return res[0]
    return res



from e2b_code_interpreter import Sandbox, CommandExitException, AsyncSandbox
from langchain_core.language_models.chat_models import BaseChatModel

from typing import Callable, Any, Literal, Optional

from openevals.code.base import (
    _create_base_code_evaluator,
    _create_async_base_code_evaluator,
)
from openevals.types import SimpleEvaluator, SimpleAsyncEvaluator

from openevals.code.e2b.sandbox.files import (
    EXTRACT_IMPORT_NAMES,
    PYTHON_EVALUATOR_SEPARATOR,
)


def _create_e2b_execution_command(
    *,
    execution_command: str = "python",
) -> str:
    return (" && ").join(
        [
            f"echo '{EXTRACT_IMPORT_NAMES}' > extract_import_names.py",
            "export PIP_DISABLE_PIP_VERSION_CHECK=1",
            "python3 extract_import_names.py > openevals_requirements.txt",
            'if command -v "uv" >/dev/null 2>&1; then uv venv --quiet && uv pip install -r openevals_requirements.txt --quiet; else pip install -r openevals_requirements.txt --quiet --upgrade-strategy only-if-needed; fi',
            f"echo '{PYTHON_EVALUATOR_SEPARATOR}'",
            f'if command -v "uv" >/dev/null 2>&1; then uv run {execution_command} outputs.py; else {execution_command} outputs.py; fi',
        ]
    )


def create_e2b_execution_evaluator(
    *,
    sandbox: Sandbox,
    environment_variables: Optional[dict[str, str]] = None,
    execution_command: str = "python",
    sandbox_project_directory: Optional[str] = None,
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    model: Optional[str] = None,
    client: Optional[BaseChatModel] = None,
) -> SimpleEvaluator:
    """Creates an evaluator that executes code in an E2B sandbox environment and returns whether
    execution was successful.

    Args:
        sandbox (Sandbox): The E2B sandbox environment for code execution.
        environment_variables (Optional[dict[str, str]], optional): Environment variables to set in the sandbox.
            Defaults to None.
        execution_command (Optional[str], optional): Command used to execute the code.
            Defaults to "python".
        sandbox_project_directory (Optional[str], optional): Directory where the code will be executed.
            Defaults to None.
        code_extraction_strategy (Literal["none", "llm", "markdown_code_blocks"], optional): Strategy for
            extracting code from the input. Defaults to "none".
        code_extractor (Optional[Callable[[Any], str]], optional): Custom function to extract code from input.
            Only used if code_extraction_strategy is "none".
        model (Optional[str], optional): The model identifier for LLM-based code extraction.
            Only used if code_extraction_strategy is "llm".
        client (Optional[BaseChatModel], optional): The chat model client for LLM-based code extraction.
            Only used if code_extraction_strategy is "llm".

    Returns:
        SimpleEvaluator: An evaluator function that returns a tuple of (success: bool, feedback: str)
            where success indicates if the code executed without errors and feedback contains the
            execution output or error messages.

    Raises:
        ValueError: If model or client are provided when code_extraction_strategy is not "llm".
    """
    if code_extraction_strategy != "llm" and (model or client):
        raise ValueError(
            "model and client may only be passed if code_extraction_strategy is 'llm'"
        )

    def _scorer(outputs: str, **kwargs: Any):
        cwd = sandbox_project_directory or "openevals"
        sandbox.files.write(f"{cwd}/outputs.py", outputs)
        try:
            cmd = _create_e2b_execution_command(execution_command=execution_command)
            sandbox.commands.run(cmd=cmd, cwd=cwd, envs=environment_variables)
            return True, None
        except CommandExitException as e:
            return False, str(e)

    return _create_base_code_evaluator(
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
        model=model,
        client=client,
        run_name="e2b_execution_evaluator",
        feedback_key="execution_succeeded",
    )


def create_async_e2b_execution_evaluator(
    *,
    sandbox: AsyncSandbox,
    environment_variables: Optional[dict[str, str]] = None,
    execution_command: str = "python",
    sandbox_project_directory: Optional[str] = None,
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    model: Optional[str] = None,
    client: Optional[BaseChatModel] = None,
) -> SimpleAsyncEvaluator:
    """Creates an async evaluator that executes code in an E2B sandbox environment and returns whether
    execution was successful.

    Args:
        sandbox (AsyncSandbox): The E2B async sandbox environment for code execution.
        environment_variables (Optional[dict[str, str]], optional): Environment variables to set in the sandbox.
            Defaults to None.
        execution_command (Optional[str], optional): Command used to execute the code.
            Defaults to "python".
        sandbox_project_directory (Optional[str], optional): Directory where the code will be executed.
            Defaults to None.
        code_extraction_strategy (Literal["none", "llm", "markdown_code_blocks"], optional): Strategy for
            extracting code from the input. Defaults to "none".
        code_extractor (Optional[Callable[[Any], str]], optional): Custom function to extract code from input.
            Only used if code_extraction_strategy is "none".
        model (Optional[str], optional): The model identifier for LLM-based code extraction.
            Only used if code_extraction_strategy is "llm".
        client (Optional[BaseChatModel], optional): The chat model client for LLM-based code extraction.
            Only used if code_extraction_strategy is "llm".

    Returns:
        SimpleAsyncEvaluator: An evaluator function that returns a tuple of (success: bool, feedback: str)
            where success indicates if the code executed without errors and feedback contains the
            execution output or error messages.

    Raises:
        ValueError: If model or client are provided when code_extraction_strategy is not "llm".
    """
    if code_extraction_strategy != "llm" and (model or client):
        raise ValueError(
            "model and client may only be passed if code_extraction_strategy is 'llm'"
        )

    async def _scorer(outputs: str, **kwargs: Any):
        cwd = sandbox_project_directory or "openevals"
        await sandbox.files.write(f"{cwd}/outputs.py", outputs)
        try:
            cmd = _create_e2b_execution_command(execution_command=execution_command)
            await sandbox.commands.run(cmd=cmd, cwd=cwd, envs=environment_variables)
            return True, None
        except CommandExitException as e:
            return False, str(e)

    return _create_async_base_code_evaluator(
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
        model=model,
        client=client,
        run_name="e2b_execution_evaluator",
        feedback_key="execution_succeeded",
    )



import pytest

from openevals.llm import create_llm_as_judge
from openevals.prompts.hallucination import HALLUCINATION_PROMPT


@pytest.mark.langsmith
def test_llm_as_judge_hallucination():
    inputs = {
        "question": "Who was the first president of the Star Republic of Oiewjoie?",
    }
    outputs = {"answer": "Bzkeoei Ahbeijo"}
    llm_as_judge = create_llm_as_judge(
        prompt=HALLUCINATION_PROMPT,
        feedback_key="hallucination",
        model="openai:o3-mini",
    )
    context = "The Star Republic of Oiewjoie is a country that exists in the universe. The first president of the Star Republic of Oiewjoie was Bzkeoei Ahbeijo."
    with pytest.raises(KeyError):
        eval_result = llm_as_judge(inputs=inputs, outputs=outputs)
    eval_result = llm_as_judge(
        inputs=inputs, outputs=outputs, context=context, reference_outputs=""
    )
    assert eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_hallucination_not_correct():
    inputs = {
        "question": "Who was the first president of the Star Republic of Oiewjoie?",
    }
    outputs = {"answer": "John Adams"}
    llm_as_judge = create_llm_as_judge(
        prompt=HALLUCINATION_PROMPT,
        feedback_key="hallucination",
        model="openai:o3-mini",
    )
    context = "The Star Republic of Oiewjoie is a country that exists in the universe. The first president of the Star Republic of Oiewjoie was Bzkeoei Ahbeijo."
    eval_result = llm_as_judge(
        inputs=inputs, outputs=outputs, context=context, reference_outputs=""
    )
    assert not eval_result["score"]



CORRECTNESS_PROMPT = """You are an expert data labeler evaluating model outputs for correctness. Your task is to assign a score based on the following rubric:

<Rubric>
  A correct answer:
  - Provides accurate and complete information
  - Contains no factual errors
  - Addresses all parts of the question
  - Is logically consistent
  - Uses precise and accurate terminology

  When scoring, you should penalize:
  - Factual errors or inaccuracies
  - Incomplete or partial answers
  - Misleading or ambiguous statements
  - Incorrect terminology
  - Logical inconsistencies
  - Missing key information
</Rubric>

<Instructions>
  - Carefully read the input and output
  - Check for factual accuracy and completeness
  - Focus on correctness of information rather than style or verbosity
</Instructions>

<Reminder>
  The goal is to evaluate factual correctness and completeness of the response.
</Reminder>

<input>
{inputs}
</input>

<output>
{outputs}
</output>

Use the reference outputs below to help you evaluate the correctness of the response:

<reference_outputs>
{reference_outputs}
</reference_outputs>
"""



import pytest

from openevals.llm import create_llm_as_judge
from openevals.prompts.correctness import CORRECTNESS_PROMPT


@pytest.mark.langsmith
def test_llm_as_judge_correctness():
    inputs = {
        "question": "Who was the first president of the United States?",
    }
    outputs = {"answer": "George Washington"}
    reference_outputs = {"answer": "George Washington"}
    llm_as_judge = create_llm_as_judge(
        prompt=CORRECTNESS_PROMPT,
        feedback_key="correctness",
        model="openai:o3-mini",
    )
    with pytest.raises(KeyError):
        eval_result = llm_as_judge(inputs=inputs, outputs=outputs)
    eval_result = llm_as_judge(
        inputs=inputs, outputs=outputs, reference_outputs=reference_outputs
    )
    assert eval_result["score"]


@pytest.mark.langsmith
def test_llm_as_judge_correctness_not_correct():
    inputs = {
        "question": "Who was the first president of the United States?",
    }
    outputs = {"answer": "John Adams"}
    reference_outputs = {"answer": "George Washington"}
    llm_as_judge = create_llm_as_judge(
        prompt=CORRECTNESS_PROMPT,
        feedback_key="correctness",
        model="openai:o3-mini",
    )
    eval_result = llm_as_judge(
        inputs=inputs, outputs=outputs, reference_outputs=reference_outputs
    )
    assert not eval_result["score"]



CODE_CORRECTNESS_PROMPT = """You are an expert code reviewer evaluating code for correctness. Your task is to assign a score based on the following rubric:

<Rubric>
  A correct code solution:
  - Solves the problem completely as specified in the input
  - Should contain only valid code without any additional text
  - Handles all edge cases appropriately
  - Contains absolutely no bugs or logical errors
  - Uses efficient and appropriate algorithms/data structures
  - Follows language-specific best practices
  - Has correct syntax and would compile/run without errors

  When scoring, you should penalize:
  - Logical errors or bugs that would cause incorrect behavior
  - Missing edge case handling
  - Overly inefficient implementations when better approaches exist
  - Incomplete solutions that don't address all requirements
  - Syntax errors that would prevent compilation/execution
  - Security vulnerabilities or unsafe practices
  - Additional text that is not code
</Rubric>

<Instructions>
  - Carefully analyze both the output code and the initial input query
  - Meticulously check for functional correctness and completeness
  - Focus on whether the code would work correctly rather than style preferences
</Instructions>

<Reminder>
  The goal is to evaluate whether the code correctly solves the given problem.
</Reminder>

<input>
{inputs}
</input>

<output>
{outputs}
</output>
"""

CODE_CORRECTNESS_PROMPT_WITH_REFERENCE_OUTPUTS = """You are an expert code reviewer evaluating code for correctness. Your task is to assign a score based on the following rubric:

<Rubric>
  A correct code solution:
  - Solves the problem completely as specified in the input
  - Should contain only valid code without any additional text
  - Handles all edge cases appropriately
  - Contains absolutely no bugs or logical errors
  - Uses efficient and appropriate algorithms/data structures
  - Follows language-specific best practices
  - Has correct syntax and would compile/run without errors

  When scoring, you should penalize:
  - Logical errors or bugs that would cause incorrect behavior
  - Missing edge case handling
  - Overly inefficient implementations when better approaches exist
  - Incomplete solutions that don't address all requirements
  - Syntax errors that would prevent compilation/execution
  - Security vulnerabilities or unsafe practices
  - Additional text that is not code
</Rubric>

<Instructions>
  - Carefully analyze both the output code and the initial input query
  - Meticulously check for functional correctness and completeness
  - Focus on whether the code would work correctly rather than style preferences
  - Compare the output with the reference output to verify correctness
  - The reference output represents the expected behavior or result
  - Code that produces results matching the reference output should be scored higher
  - Consider edge cases where the code might produce correct results for the given examples but fail in other scenarios
</Instructions>

<Reminder>
  The goal is to evaluate whether the code correctly solves the given problem and produces output that matches the reference.
</Reminder>

<input>
{inputs}
</input>

<output>
{outputs}
</output>

<reference_output>
{reference_outputs}
</reference_output>
"""



from __future__ import annotations
from openevals.utils import (
    _run_evaluator,
    _arun_evaluator,
    _convert_to_openai_message,
    _normalize_to_openai_messages_list,
)
from openevals.types import (
    EvaluatorResult,
    SimpleEvaluator,
    SimpleAsyncEvaluator,
    RunnableLike,
    ModelClient,
    ChatCompletionMessage,
    FewShotExample,
    ScoreType,
)

from langchain.chat_models import init_chat_model
from langchain_core.language_models.chat_models import BaseChatModel, BaseMessage
from langsmith import traceable

from typing import (
    Any,
    Awaitable,
    Callable,
    Optional,
    Union,
)

import json


def _append_few_shot_examples(
    *,
    messages: list[ChatCompletionMessage],
    few_shot_examples: list[FewShotExample],
) -> list[ChatCompletionMessage]:
    # Find the last user message to append examples to
    last_user_message_idx = None
    for i, msg in enumerate(messages[::-1]):
        if msg.get("role") == "user":
            last_user_message_idx = len(messages) - 1 - i
            break

    if last_user_message_idx is None:
        raise ValueError(
            "Appending few-shot examples requires a user message in the provided prompt"
        )

    messages[last_user_message_idx]["content"] += "\n\n" + "\n".join(  # type: ignore
        [
            f"<example>\n<input>{example['inputs']}</input>\n<output>{example['outputs']}</output>"
            + (
                f"\n<reasoning>{example['reasoning']}</reasoning>"
                if "reasoning" in example
                else ""
            )
            + (f"\n<score>{example['score']}</score>" if "score" in example else "")
            + "\n</example>"
            for example in few_shot_examples
        ]
    )
    return messages


def _construct_output_schema(
    *,
    schema: Optional[dict] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
) -> tuple[dict, str]:
    json_schema = (
        schema
        if schema is not None
        else {
            "type": "object",
            "additionalProperties": False,
        }
    )
    # Set the description for the score schema
    if choices:
        description = "A number that represents the degree to which the criteria in the prompt are met."
        score_schema = {
            "type": "number",
            "description": description,
            "enum": choices,
        }
    elif continuous:
        description = "A number that represents the degree to which the criteria in the prompt are met, from 0.0 to 1.0. 1.0 means the criteria are met perfectly. 0.0 means none of the criteria are met, 0.5 means exactly half of the criteria are met."
        score_schema = {
            "type": "number",
            "description": description,
        }
    else:
        description = "A score that is true if criteria in the prompt are met, and false otherwise."
        score_schema = {
            "type": "boolean",
            "description": description,
        }

    # Add reasoning if passed
    if schema is None:
        if use_reasoning:
            json_schema["properties"] = {
                "reasoning": {
                    "type": "string",
                    "description": "A human-readable explanation of the score. You MUST end the reasoning with a sentence that says: Thus, the score should be: SCORE_YOU_ASSIGN.",
                },
                "score": score_schema,
            }
            json_schema["required"] = ["reasoning", "score"]
        else:
            json_schema["properties"] = {
                "score": score_schema,
            }
            json_schema["required"] = ["score"]

    return (json_schema, description)


def _stringify_prompt_param(param: Any) -> Any:
    if isinstance(param, str):
        return param
    elif isinstance(param, BaseMessage):
        return json.dumps(_convert_to_openai_message(param))
    elif isinstance(param, dict):
        if "messages" in param and isinstance(param["messages"], list):
            param["messages"] = [
                _convert_to_openai_message(msg) if isinstance(msg, BaseMessage) else msg
                for msg in param["messages"]
            ]
            return json.dumps(param)
        else:
            return json.dumps(param)
    elif isinstance(param, list):
        return json.dumps(
            [
                _convert_to_openai_message(msg) if isinstance(msg, BaseMessage) else msg
                for msg in param
            ]
        )
    else:
        return json.dumps(param)


def _create_llm_as_judge_scorer(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    system: Optional[str] = None,
    schema: Optional[dict] = None,
    judge: Optional[
        Union[
            ModelClient,
            BaseChatModel,
        ]
    ] = None,
    model: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> Callable[..., ScoreType]:
    def get_score(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Optional[Union[str, dict]] = None,
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ):
        if system is not None and not isinstance(prompt, str):
            raise ValueError(
                "`system` is only supported when `prompt` is a string template"
            )

        stringified_outputs = (
            _stringify_prompt_param(outputs) if outputs is not None else None
        )
        stringified_reference_outputs = (
            _stringify_prompt_param(reference_outputs)
            if reference_outputs is not None
            else None
        )
        stringified_inputs = (
            _stringify_prompt_param(inputs) if inputs is not None else None
        )

        prompt_params = {
            "inputs": stringified_inputs,
            "outputs": stringified_outputs,
            "reference_outputs": stringified_reference_outputs,
            **kwargs,
        }
        filtered_prompt_params = {
            k: v for k, v in prompt_params.items() if v is not None
        }

        if isinstance(prompt, RunnableLike):
            formatted_prompt = prompt.invoke(filtered_prompt_params)
            messages = _normalize_to_openai_messages_list(formatted_prompt.messages)
        elif isinstance(prompt, str):
            formatted_prompt = prompt.format(**filtered_prompt_params)
            messages = [
                {"role": "user", "content": formatted_prompt},
            ]
        else:
            messages = prompt(
                inputs=inputs,
                outputs=outputs,
                reference_outputs=reference_outputs,
                **kwargs,
            )

        if system is not None:
            messages = [
                {"role": "system", "content": system},
                *messages,
            ]

        # Add few shot examples to the prompt
        if few_shot_examples:
            messages = _append_few_shot_examples(
                messages=messages,
                few_shot_examples=few_shot_examples,
            )

        (json_schema, description) = _construct_output_schema(
            schema=schema,
            continuous=continuous,
            choices=choices,
            use_reasoning=use_reasoning,
        )

        nonlocal judge
        nonlocal model

        if judge is None:
            if model is None:
                raise ValueError("a `model` string is required (e.g. 'openai:o3-mini')")
            judge = init_chat_model(model=model)

        if isinstance(judge, BaseChatModel):
            response = judge.with_structured_output(
                {
                    "title": "score",
                    "description": description,
                    **json_schema,
                }
            ).invoke(messages)  # type: ignore
            if schema is None:
                if use_reasoning:
                    return (response["score"], response["reasoning"])  # type: ignore
                return response["score"]  # type: ignore
            else:
                return response
        elif isinstance(judge, ModelClient):
            if model is None:
                raise ValueError("a `model` string is required (e.g. 'openai:o3-mini')")
            if model.startswith("openai:"):
                model = model[len("openai:") :]
            params = {
                "messages": messages,
                "model": model,
                "response_format": {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "score",
                        "strict": True,
                        "schema": json_schema,
                    },
                },
            }

            @traceable(
                run_type="llm",
                metadata={
                    "ls_provider": "openai",
                    "ls_model_name": model,
                    "ls_model_type": "chat",
                },
            )
            def invoke_llm(**params):
                return judge.chat.completions.create(**params)

            response = invoke_llm(**params)  # type: ignore
            parsed = json.loads(response.choices[0].message.content)  # type: ignore
            if schema is None:
                if use_reasoning:
                    return (parsed["score"], parsed["reasoning"])  # type: ignore
                return parsed["score"]  # type: ignore
            else:
                return parsed
        else:
            raise ValueError("`judge` must be a ModelClient or BaseChatModel")

    return get_score


def _create_async_llm_as_judge_scorer(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    system: Optional[str] = None,
    schema: Optional[dict] = None,
    judge: Optional[
        Union[
            ModelClient,
            BaseChatModel,
        ]
    ] = None,
    model: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> Callable[..., Union[Awaitable[ScoreType], ScoreType]]:
    async def aget_score(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Optional[Union[str, dict]] = None,
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ):
        if system is not None and not isinstance(prompt, str):
            raise ValueError(
                "`system` is only supported when `prompt` is a string template"
            )

        stringified_outputs = (
            _stringify_prompt_param(outputs) if outputs is not None else None
        )
        stringified_reference_outputs = (
            _stringify_prompt_param(reference_outputs)
            if reference_outputs is not None
            else None
        )
        stringified_inputs = (
            _stringify_prompt_param(inputs) if inputs is not None else None
        )

        prompt_params = {
            "inputs": stringified_inputs,
            "outputs": stringified_outputs,
            "reference_outputs": stringified_reference_outputs,
            **kwargs,
        }
        filtered_prompt_params = {
            k: v for k, v in prompt_params.items() if v is not None
        }

        if isinstance(prompt, RunnableLike):
            formatted_prompt = await prompt.ainvoke(filtered_prompt_params)
            messages = _normalize_to_openai_messages_list(formatted_prompt.messages)
        elif isinstance(prompt, str):
            formatted_prompt = prompt.format(**filtered_prompt_params)
            messages = [
                {"role": "user", "content": formatted_prompt},
            ]
        else:
            messages = prompt(
                inputs=inputs,
                outputs=outputs,
                reference_outputs=reference_outputs,
                **kwargs,
            )

        if system is not None:
            messages = [
                {"role": "system", "content": system},
                *messages,
            ]

        # Add few shot examples to the prompt
        if few_shot_examples:
            messages = _append_few_shot_examples(
                messages=messages,
                few_shot_examples=few_shot_examples,
            )

        (json_schema, description) = _construct_output_schema(
            schema=schema,
            continuous=continuous,
            choices=choices,
            use_reasoning=use_reasoning,
        )

        nonlocal judge
        nonlocal model

        if judge is None:
            if model is None:
                raise ValueError("a `model` string is required (e.g. 'openai:o3-mini')")
            judge = init_chat_model(model=model)

        if isinstance(judge, BaseChatModel):
            response = await judge.with_structured_output(
                {
                    "title": "score",
                    "description": description,
                    **json_schema,
                }
            ).ainvoke(messages)  # type: ignore
            if schema is None:
                if use_reasoning:
                    return (response["score"], response["reasoning"])  # type: ignore
                return response["score"]  # type: ignore
            else:
                return response
        elif isinstance(judge, ModelClient):
            if model is None:
                raise ValueError("a `model` string is required (e.g. 'openai:o3-mini')")
            if model.startswith("openai:"):
                model = model[len("openai:") :]
            params = {
                "messages": messages,
                "model": model,
                "response_format": {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "score",
                        "strict": True,
                        "schema": json_schema,
                    },
                },
            }

            @traceable(
                run_type="llm",
                metadata={
                    "ls_provider": "openai",
                    "ls_model_name": model,
                    "ls_model_type": "chat",
                },
            )
            async def ainvoke_llm(**params):
                return await judge.chat.completions.create(**params)

            response = await ainvoke_llm(**params)  # type: ignore
            parsed = json.loads(response.choices[0].message.content)  # type: ignore
            if schema is None:
                if use_reasoning:
                    return (parsed["score"], parsed["reasoning"])  # type: ignore
                return parsed["score"]  # type: ignore
            else:
                return parsed
        else:
            raise ValueError("`judge` must be a ModelClient or BaseChatModel")

    return aget_score


def create_llm_as_judge(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    feedback_key: str = "score",
    judge: Optional[Union[ModelClient, BaseChatModel]] = None,
    model: Optional[str] = None,
    system: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> SimpleEvaluator:
    """Create an evaluator that uses an LLM to assess output quality based on specified criteria.
    Handles prompt formatting, LLM invocation, and structured output parsing.

    Args:
        prompt: The evaluation prompt, can be a string template, LangChain prompt template, or callable
            that returns a list of chat messages.
        feedback_key: Key used to store the evaluation result, defaults to "score".
        judge: The LLM used for evaluation. Can be an OpenAI client),
            or a BaseChatModel. If an OpenAI client, must specify "model" as well.
            If omitted, "model" will be used to instantiate a LangChain model instance
            by model string.
        model: Model identifier to use. If "judge" is an OpenAI client,
            this argument should be a model name directly. If "judge" is omitted, must be a valid
            LangChain model identifier. See `init_chat_model` docs for more details:
            https://python.langchain.com/docs/how_to/chat_models_universal_init/.
        system: Optional system message to prepend to the prompt.
        continuous: If True, score will be a float between 0 and 1. If False, score will be boolean. Defaults to False.
        choices: Optional list of specific float values the score must be chosen from.
        use_reasoning: If True, includes explanation for the score in the output. Defaults to True.
        few_shot_examples: Optional list of example evaluations to append to the prompt.

    Returns:
        A function that takes inputs, outputs, reference_outputs, and other kwargs, formats them into
        a prompt, invokes the judge, and returns an evaluation result.

    Example:
        ```python
        from openevals.llm import create_llm_as_judge

        evaluator = create_llm_as_judge(
            prompt="Rate the quality of this response from 0 to 1: {outputs}",
            continuous=True
        )
        result = evaluator(
            inputs={"question": "What color is the sky?"},
            outputs={"response": "Blue"},
        )
        ```
    """
    scorer = _create_llm_as_judge_scorer(
        prompt=prompt,
        judge=judge,
        system=system,
        model=model,
        continuous=continuous,
        choices=choices,
        use_reasoning=use_reasoning,
        few_shot_examples=few_shot_examples,
    )

    def _wrapped_evaluator(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Optional[Union[str, dict]] = None,
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ) -> EvaluatorResult:
        run_name = (
            "llm_as_judge"
            if feedback_key == "score"
            else f"llm_as_{feedback_key}_judge"
        )
        res = _run_evaluator(
            run_name=run_name,
            scorer=scorer,
            feedback_key=feedback_key,
            inputs=inputs,
            outputs=outputs,
            reference_outputs=reference_outputs,
            **kwargs,
        )
        if isinstance(res, list):
            return res[0]
        return res  # type: ignore

    return _wrapped_evaluator  # type: ignore


def create_async_llm_as_judge(
    *,
    prompt: str | RunnableLike | Callable[..., list[ChatCompletionMessage]],
    feedback_key: str = "score",
    judge: Optional[Union[ModelClient, BaseChatModel]] = None,
    model: Optional[str] = None,
    system: Optional[str] = None,
    continuous: bool = False,
    choices: Optional[list[float]] = None,
    use_reasoning: bool = True,
    few_shot_examples: Optional[list[FewShotExample]] = None,
) -> SimpleAsyncEvaluator:
    """Create an evaluator that uses an LLM to assess output quality based on specified criteria.
    Handles prompt formatting, LLM invocation, and structured output parsing.

    Args:
        prompt: The evaluation prompt, can be a string template, LangChain prompt template, or callable
            that returns a list of chat messages.
        feedback_key: Key used to store the evaluation result, defaults to "score".
        judge: The LLM used for evaluation. Can be an OpenAI client),
            or a BaseChatModel. If an OpenAI client, must specify "model" as well.
            If omitted, "model" will be used to instantiate a LangChain model instance
            by model string.
        model: Model identifier to use. If "judge" is an OpenAI client,
            this argument should be a model name directly. If "judge" is omitted, must be a valid
            LangChain model identifier. See `init_chat_model` docs for more details:
            https://python.langchain.com/docs/how_to/chat_models_universal_init/.
        system: Optional system message to prepend to the prompt.
        continuous: If True, score will be a float between 0 and 1. If False, score will be boolean. Defaults to False.
        choices: Optional list of specific float values the score must be chosen from.
        use_reasoning: If True, includes explanation for the score in the output. Defaults to True.
        few_shot_examples: Optional list of example evaluations to append to the prompt.

    Returns:
        A function that takes inputs, outputs, reference_outputs, and other kwargs, formats them into
        a prompt, invokes the judge, and returns an evaluation result.

    Example:
        ```python
        from openevals.llm import create_async_llm_as_judge

        evaluator = create_async_llm_as_judge(
            prompt="Rate the quality of this response from 0 to 1: {outputs}",
            continuous=True
        )
        result = await evaluator(
            inputs={"question": "What color is the sky?"},
            outputs={"response": "Blue"},
        )
        ```
    """
    scorer = _create_async_llm_as_judge_scorer(
        prompt=prompt,
        judge=judge,
        system=system,
        model=model,
        continuous=continuous,
        choices=choices,
        use_reasoning=use_reasoning,
        few_shot_examples=few_shot_examples,
    )

    async def _wrapped_evaluator(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Optional[Union[str, dict]] = None,
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ) -> EvaluatorResult:
        run_name = (
            "llm_as_judge"
            if feedback_key == "score"
            else f"llm_as_{feedback_key}_judge"
        )
        res = await _arun_evaluator(
            run_name=run_name,
            scorer=scorer,
            feedback_key=feedback_key,
            inputs=inputs,
            outputs=outputs,
            reference_outputs=reference_outputs,
            **kwargs,
        )
        if isinstance(res, list):
            return res[0]
        return res

    return _wrapped_evaluator  # type: ignore



import asyncio
import json
import subprocess
import tempfile
import os

from typing import Optional, Callable, Any, Union, Literal, Awaitable, Tuple

from openevals.code.llm import (
    _create_base_code_evaluator,
    _create_async_base_code_evaluator,
)
from openevals.types import SimpleEvaluator, SimpleAsyncEvaluator

from langchain_core.language_models.chat_models import BaseChatModel


def _parse_mypy_output(stdout: bytes) -> Tuple[bool, str]:
    try:
        # Parse the output - mypy gives line-by-line error messages
        errors = []
        for line in stdout.decode().split("\n"):
            line_components = line.strip().split(":")
            if len(line_components) > 2 and line_components[0].endswith(".py"):
                display_line = ":".join(["python_file.py", *line_components])
                errors.append(display_line)
        score = len(errors) == 0
        return (score, "\n".join(errors))
    except json.JSONDecodeError:
        return (False, f"Failed to parse Mypy output: {stdout.decode()}")


def _analyze_with_mypy(
    *,
    output: str,
    mypy_cli_args: list[str],
) -> Tuple[bool, str]:
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".py", mode="w", delete=False) as temp:
        temp.write(output)
        temp_path = temp.name

    try:
        # Run mypy with specific flags
        result = subprocess.run(
            [
                "mypy",
                *mypy_cli_args,
                temp_path,
            ],
            capture_output=True,
        )
        return _parse_mypy_output(result.stdout)  # type: ignore

    finally:
        # Clean up
        os.unlink(temp_path)


async def _analyze_with_mypy_async(
    *,
    output: str,
    mypy_cli_args: list[str],
) -> Tuple[bool, str]:
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".py", mode="w", delete=False) as temp:
        temp.write(output)
        temp_path = temp.name

    try:
        # Run mypy with specific flags asynchronously
        process = await asyncio.create_subprocess_exec(
            "mypy",
            *(mypy_cli_args or []),
            temp_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, _ = await process.communicate()

        return _parse_mypy_output(stdout)  # type: ignore

    finally:
        # Clean up
        os.unlink(temp_path)


def create_mypy_evaluator(
    *,
    mypy_cli_args: list[str] = [
        "--no-incremental",
        "--disallow-untyped-calls",
        "--disallow-incomplete-defs",
        "--ignore-missing-imports",
    ],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    client: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
) -> SimpleEvaluator:
    """Create an evaluator that checks Python code with mypy.

    This function creates an evaluator that runs mypy on Python code
    to check for type errors and other issues. It supports different strategies
    for extracting code from the input.

    Args:
        mypy_cli_args: Command line arguments to pass to mypy. Defaults to a set of
            common type checking flags.
        code_extraction_strategy: Strategy for extracting code from the input.
            Options are:
            - "none": Use the input as-is
            - "llm": Use an LLM to extract code (requires client or model)
            - "markdown_code_blocks": Extract code from markdown code blocks
        code_extractor: Optional custom function to extract code from the input.
        client: LLM client to use for code extraction if strategy is "llm".
        model: Model identifier to use for code extraction if strategy is "llm".

    Returns:
        An evaluator function that runs mypy on the input code.

    Raises:
        ValueError: If client or model is provided with a code_extraction_strategy
            other than "llm".
    """
    if code_extraction_strategy != "llm" and (client or model):
        raise ValueError(
            "client and model may only be passed if code_extraction_strategy is 'llm'"
        )

    def _scorer(
        *,
        outputs: str,
        **kwargs,
    ):
        return _analyze_with_mypy(
            output=outputs,
            mypy_cli_args=mypy_cli_args,
        )

    return _create_base_code_evaluator(
        model=model,
        client=client,
        run_name="mypy_evaluator",
        feedback_key="mypy_succeeded",
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )


def create_async_mypy_evaluator(
    *,
    mypy_cli_args: list[str] = [
        "--no-incremental",
        "--disallow-untyped-calls",
        "--disallow-incomplete-defs",
        "--ignore-missing-imports",
    ],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], Union[str, Awaitable[str]]]] = None,
    client: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
) -> SimpleAsyncEvaluator:
    """Create an asynchronous evaluator that checks Python code with mypy.

    This function creates an asynchronous evaluator that runs mypy on Python code
    to check for type errors and other issues. It supports different strategies
    for extracting code from the input.

    Args:
        mypy_cli_args: Command line arguments to pass to mypy. Defaults to a set of
            common type checking flags.
        code_extraction_strategy: Strategy for extracting code from the input.
            Options are:
            - "none": Use the input as-is
            - "llm": Use an LLM to extract code (requires client or model)
            - "markdown_code_blocks": Extract code from markdown code blocks
        code_extractor: Optional custom function to extract code from the input.
        client: LLM client to use for code extraction if strategy is "llm".
        model: Model identifier to use for code extraction if strategy is "llm".

    Returns:
        An asynchronous evaluator function that runs mypy on the input code.

    Raises:
        ValueError: If client or model is provided with a code_extraction_strategy
            other than "llm".
    """
    if code_extraction_strategy != "llm" and (client or model):
        raise ValueError(
            "client and model may only be passed if code_extraction_strategy is 'llm'"
        )

    async def _scorer(
        *,
        outputs: str,
        **kwargs,
    ):
        return await _analyze_with_mypy_async(
            output=outputs,
            mypy_cli_args=mypy_cli_args,
        )

    return _create_async_base_code_evaluator(
        model=model,
        client=client,
        run_name="mypy_check",
        feedback_key="mypy_succeeded",
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )



from openevals.json import create_json_match_evaluator
from langsmith import Client
import pytest


@pytest.mark.langsmith
def test_json_match_base():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 2}
    evaluator = create_json_match_evaluator()
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert len(result) == 2
    assert result[0]["key"] == "json_match:a"
    assert result[0]["score"] == 1.0
    assert result[1]["key"] == "json_match:b"
    assert result[1]["score"] == 1.0


@pytest.mark.langsmith
def test_json_match_mix():
    outputs = {"a": "Mango, Bananas", "b": 2}
    reference_outputs = {"a": "Bananas, Mango", "b": 3}
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        rubric={"a": "Does the answer mention all the fruits in the reference answer?"},
        aggregator="average",
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0.5


@pytest.mark.langsmith
def test_json_match_average():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    evaluator = create_json_match_evaluator(aggregator="average")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0.5


@pytest.mark.langsmith
def test_json_match_exclude():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    evaluator = create_json_match_evaluator(aggregator="average", exclude_keys=["b"])
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_all():
    outputs = {"a": 1, "b": 2}
    reference_outputs = {"a": 1, "b": 3}
    evaluator = create_json_match_evaluator(aggregator="all")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_rubric():
    outputs = {
        "name": "Harrison Chase",
        "description": "CEO of LangChain, used to work at Kensho + Robust Intelligence.",
    }
    reference_outputs = {
        "name": "Harrison Chase",
        "description": "Harrison chase is the CEO of LangChain. He used to work at Kensho and Robust Intelligence.",
    }
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        aggregator="all",
        rubric={
            "description": "Is the correct job title and company mentioned, as well as previous companies?"
        },
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_rubric_wrong():
    outputs = {
        "name": "Harrison Chase",
        "description": "CEO of LangChain, used to work at Kensho.",
    }
    reference_outputs = {
        "name": "Harrison Chase",
        "description": "Harrison chase is the CEO of LangChain. He used to work at Kensho and Robust Intelligence.",
    }
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        aggregator="all",
        rubric={
            "description": "Is the correct job title and company mentioned, as well as previous companies?"
        },
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_rubric_with_reasoning():
    outputs = {"description": "CEO of LangChain, used to work at Kensho."}
    reference_outputs = {
        "description": "Harrison chase is the CEO of LangChain. He used to work at Kensho and Robust Intelligence."
    }
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        rubric={
            "description": "Is the correct job title and company mentioned, as well as previous companies?"
        },
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:description"
    assert result[0]["score"] == 0
    assert result[0]["comment"] is not None


@pytest.mark.langsmith
def test_json_match_rubric_without_reasoning():
    outputs = {"description": "CEO of LangChain, used to work at Kensho."}
    reference_outputs = {
        "description": "Harrison chase is the CEO of LangChain. He used to work at Kensho and Robust Intelligence."
    }
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        aggregator="all",
        rubric={
            "description": "Is the correct job title and company mentioned, as well as previous companies?"
        },
        use_reasoning=False,
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0
    assert result[0]["comment"] is None


@pytest.mark.langsmith
def test_json_match_rubric_with_reasoning_individual_key():
    outputs = {
        "name": "Harrison Chase",
        "description": "CEO of LangChain, used to work at Kensho.",
    }
    reference_outputs = {
        "name": "Harrison Chase",
        "description": "Harrison chase is the CEO of LangChain. He used to work at Kensho and Robust Intelligence.",
    }
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        rubric={
            "description": "Is the correct job title and company mentioned, as well as previous companies?"
        },
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert len(result) == 2
    assert result[0]["key"] == "json_match:name"
    assert result[0]["score"] == 1.0
    assert result[1]["key"] == "json_match:description"
    assert result[1]["score"] == 0
    assert result[1]["comment"] is not None


@pytest.mark.langsmith
def test_json_match_list_all_none():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    evaluator = create_json_match_evaluator()
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    result = sorted(result, key=lambda x: x["key"])
    assert len(result) == 2
    assert result[0]["key"] == "json_match:a"
    assert result[0]["score"] == 1.0
    assert result[1]["key"] == "json_match:b"
    assert result[1]["score"] == 1.0


@pytest.mark.langsmith
def test_json_match_list_average_none():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 3},
    ]
    evaluator = create_json_match_evaluator(list_aggregator="average")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    result = sorted(result, key=lambda x: x["key"])
    assert len(result) == 2
    assert result[0]["key"] == "json_match:a"
    assert result[0]["score"] == 1.0
    assert result[1]["key"] == "json_match:b"
    assert result[1]["score"] == 0.5


@pytest.mark.langsmith
def test_json_match_list_all_all():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    evaluator = create_json_match_evaluator(aggregator="all")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_list_average_all():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 3},
    ]
    evaluator = create_json_match_evaluator(list_aggregator="average", aggregator="all")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0.5


@pytest.mark.langsmith
def test_json_match_list_all_average():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 2, "b": 2},
    ]
    evaluator = create_json_match_evaluator(aggregator="average")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_average_average():
    outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 3},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0.75


@pytest.mark.langsmith
def test_json_match_list_mismatch_all_none():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator()
    results = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    results = sorted(results, key=lambda x: x["key"])
    assert len(results) == 4
    assert results[0]["key"] == "json_match:a"
    assert results[0]["score"] == 1.0
    assert results[1]["key"] == "json_match:b"
    assert results[1]["score"] == 0
    assert results[2]["key"] == "json_match:c"
    assert results[2]["score"] == 0
    assert results[3]["key"] == "json_match:d"
    assert results[3]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_mismatch_average_none():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(list_aggregator="average")
    results = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    results = sorted(results, key=lambda x: x["key"])
    assert len(results) == 4
    assert results[0]["key"] == "json_match:a"
    assert results[0]["score"] == 1.0
    assert results[1]["key"] == "json_match:b"
    assert results[1]["score"] == 0.5
    assert results[2]["key"] == "json_match:c"
    assert results[2]["score"] == 0
    assert results[3]["key"] == "json_match:d"
    assert results[3]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_mismatch_all_all():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(aggregator="all")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_mismatch_average_all():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(list_aggregator="average", aggregator="all")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:all"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_mismatch_all_average():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(aggregator="average")
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_list_mismatch_average_average():
    outputs = [
        {"a": 1, "d": 2},
        {"a": 1, "b": 2},
    ]
    reference_outputs = [
        {"a": 1, "b": 2},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0.5


@pytest.mark.langsmith
def test_json_match_list_rubric():
    outputs = [{"a": "Strawberries, Melons, Bananas"}]
    reference_outputs = [{"a": "Bananas, Strawberries, Melons"}]
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        rubric={"a": "Does the answer mention all the fruits in the reference answer?"},
        list_aggregator="average",
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:a"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_list_mismatch_output_missing():
    outputs = [
        {"a": 1, "b": 2, "d": 3},
        {"a": 1, "b": 2, "c": 3},
        {"a": 1, "b": 2, "d": 3},
    ]
    reference_outputs = [
        {"a": 1, "b": 2, "d": 3},
        {"a": 1, "b": 2, "c": 3},
        {"a": 1, "b": 2, "c": 3},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 5 / 6


@pytest.mark.langsmith
def test_json_match_mode_exact_extra_reference():
    outputs = [{"a": 1}, {"a": 1}]
    reference_outputs = [{"a": 1}, {"a": 1}, {"a": 1}]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 2 / 3


@pytest.mark.langsmith
def test_json_match_mode_exact_extra_output():
    outputs = [{"a": 1}, {"a": 1}, {"a": 1}]
    reference_outputs = [
        {"a": 1},
        {"a": 1},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 2 / 3


@pytest.mark.langsmith
def test_json_match_mode_exact_unordered():
    outputs = [{"a": 1, "d": 2, "e": 2}, {"b": 1}, {"c": 1}]
    reference_outputs = [{"b": 1, "d": 2, "e": 2}, {"a": 1}, {"c": 1}]
    evaluator = create_json_match_evaluator(
        list_aggregator="average",
        aggregator="average",
        exclude_keys=["d", "e"],
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_mode_subset_outputs():
    outputs = [{"a": 1}, {"b": 1}, {"c": 1}]
    reference_outputs = [
        {"b": 1},
        {"a": 1},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average", list_match_mode="superset"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_mode_subset_reference():
    outputs = [
        {"a": 1},
        {"b": 1},
    ]
    reference_outputs = [{"b": 1}, {"c": 1}, {"a": 1}]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average", list_match_mode="subset"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_mode_order_wrong():
    outputs = [
        {"a": 1},
        {"b": 1},
    ]
    reference_outputs = [{"b": 1}, {"a": 1}]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average", list_match_mode="ordered"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 0


@pytest.mark.langsmith
def test_json_match_mode_order_right():
    outputs = [
        {"a": 1},
        {"b": 1},
    ]
    reference_outputs = [{"a": 1}, {"b": 1}]
    evaluator = create_json_match_evaluator(
        list_aggregator="average", aggregator="average", list_match_mode="ordered"
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_mode_order_wrong_language():
    outputs = [
        {"a": 1, "c": "The Dog"},
        {"b": 1},
    ]
    reference_outputs = [{"a": 1, "c": "El Perro"}, {"b": 1}]
    evaluator = create_json_match_evaluator(
        model="openai:o3-mini",
        list_aggregator="average",
        aggregator="average",
        list_match_mode="ordered",
        rubric={
            "c": "Are the answers the same meaning, even if they are different languages?"
        },
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 1


@pytest.mark.langsmith
def test_json_match_mode_order():
    outputs = [
        {"a": 1},
        {"b": 1},
        {"c": 1},
    ]
    reference_outputs = [
        {"a": 1},
        {"b": 1},
        {"d": 1},
    ]
    evaluator = create_json_match_evaluator(
        list_aggregator="average",
        aggregator="average",
        list_match_mode="ordered",
    )
    result = evaluator(outputs=outputs, reference_outputs=reference_outputs)
    assert result[0]["key"] == "json_match:average"
    assert result[0]["score"] == 2 / 3


@pytest.mark.langsmith
def test_works_with_evaluate():
    client = Client()
    evaluator = create_json_match_evaluator()
    res = client.evaluate(lambda x: x, data="json", evaluators=[evaluator])
    for r in res:
        assert r["evaluation_results"]["results"][0].score is not None


@pytest.mark.langsmith
def test_error_no_rubric():
    with pytest.raises(ValueError):
        create_json_match_evaluator(model="openai:o3-mini")


@pytest.mark.langsmith
def test_error_no_model():
    with pytest.raises(ValueError):
        create_json_match_evaluator(rubric={"a": "foo"})



from .exact import exact_match, exact_match_async
from .llm import create_llm_as_judge, create_async_llm_as_judge

__all__ = [
    "exact_match",
    "exact_match_async",
    "create_llm_as_judge",
    "create_async_llm_as_judge",
]



from openevals.types import (
    ScoreType,
    SimpleEvaluator,
    SimpleAsyncEvaluator,
)

from openevals.utils import (
    _normalize_final_app_outputs_as_string,
    _run_evaluator,
    _arun_evaluator,
)

from typing import Any, Literal, Union, Optional, Callable, Awaitable
from typing_extensions import TypedDict

from langchain_core.language_models.chat_models import BaseChatModel
from langchain.chat_models import init_chat_model

__all__ = [
    "_create_base_code_evaluator",
    "_create_async_base_code_evaluator",
]

LLM_EXTRACTION_SYSTEM_PROMPT = """
You are an expert software auditor.

<Instructions>
  Your job is to extract code from a given text.

  - If there is code - extract it into a single script by calling the provided "ExtractCode" tool.
  - If there is no code to extract - call "NoCode".

  If you extract code, your response will be passed DIRECTLY into a code execution sandbox for further testing,
  so make sure to extract all code **without modifications**, even if it contains errors,
  since any modifications will ruin the integrity of the testing process.
  Omit installation instructions and shell commands from any code you extract.
</Instructions>
"""

LLM_EXTRACTION_USER_PROMPT = """
Extract code from the following:

<text>
{outputs}
</text>
"""


class ExtractCode(TypedDict):
    """Tool to call if there is code to extract.
    Omit installation instructions and shell commands."""

    code: str


class NoCode(TypedDict):
    """Tool to call to indicate no code was found."""

    no_code: bool


def _extract_code_from_markdown_code_blocks(text: str) -> Optional[str]:
    """
    Extract code from markdown code blocks in the provided text.

    Supports both triple backtick code blocks with or without language specifiers.

    Args:
        text: The text containing markdown code blocks

    Returns:
        A string containing only the code extracted from code blocks, with blocks
        separated by newlines
    """
    import re

    # Pattern to match code blocks with or without language specifier
    # (?s) enables dot to match newlines
    # (?:```(?:\w+)?\n(.*?)```) matches code blocks with optional language specifier
    pattern = r"(?m)^(?<!`)\`\`\`(\w*)\n([\s\S]*?)^(?<!`)\`\`\`$"

    # Find all code blocks
    matches = re.finditer(pattern, text, re.MULTILINE)

    # Filter out bash/shell blocks and collect valid code blocks
    excluded_langs = {
        "bash",
        "sh",
        "shell",
        "zsh",
        "fish",
        "console",
        "terminal",
        "json",
    }
    code_blocks = []
    for match in matches:
        lang = match.group(1).strip()
        if lang not in excluded_langs:
            code_blocks.append(match.group(2))

    if not code_blocks:
        return None  # Return None if no code blocks found

    # Join all code blocks with newlines
    return "\n".join(code_blocks)


def _create_base_code_evaluator(
    *,
    scorer: Callable[..., Union[ScoreType, tuple[bool, Optional[str], Optional[dict]]]],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    model: Optional[str] = None,
    client: Optional[BaseChatModel] = None,
    run_name: str,
    feedback_key: str,
) -> SimpleEvaluator:
    if code_extractor is not None and code_extraction_strategy != "none":
        raise ValueError(
            "`code_extractor` and `code_extraction_strategy` cannot both be provided"
        )
    if code_extraction_strategy == "llm":
        if model is None and client is None:
            raise ValueError("You must provide either a `model` string or a `client`")
        if client is None:
            client = init_chat_model(model)  # type: ignore

    def _wrapped_evaluator(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Union[str, dict],
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ):
        def _score_wrapper(*, outputs: Union[str, dict], **kwargs):
            if code_extractor is None:
                normalized_outputs = _normalize_final_app_outputs_as_string(outputs)
                if code_extraction_strategy == "llm":
                    model_with_tools = client.bind_tools([ExtractCode, NoCode])  # type: ignore
                    res = model_with_tools.invoke(
                        [
                            {"role": "system", "content": LLM_EXTRACTION_SYSTEM_PROMPT},
                            {
                                "role": "user",
                                "content": LLM_EXTRACTION_USER_PROMPT.format(
                                    outputs=normalized_outputs
                                ),
                            },
                        ],
                        {"run_name": "extract_code"},
                    )
                    if res.tool_calls[0]["name"] == "ExtractCode":  # type: ignore
                        normalized_outputs = res.tool_calls[0]["args"]["code"]  # type: ignore
                    else:
                        return (False, None, {"code_extraction_failed": True})
                elif code_extraction_strategy == "markdown_code_blocks":
                    normalized_outputs = _extract_code_from_markdown_code_blocks(  # type: ignore
                        normalized_outputs
                    )
                    if normalized_outputs is None:
                        return (False, None, {"code_extraction_failed": True})
                else:
                    # Nothing to do to extract code
                    pass
            else:
                normalized_outputs = code_extractor(outputs)
            return scorer(
                outputs=normalized_outputs,
                **kwargs,
            )

        return _run_evaluator(
            run_name=run_name,
            scorer=_score_wrapper,
            feedback_key=feedback_key,
            inputs=inputs,
            outputs=outputs,
            reference_outputs=reference_outputs,
            **kwargs,
        )

    return _wrapped_evaluator


def _create_async_base_code_evaluator(
    *,
    scorer: Callable[..., Union[ScoreType, Awaitable[ScoreType]]],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], Union[str, Awaitable[str]]]] = None,
    model: Optional[str] = None,
    client: Optional[BaseChatModel] = None,
    run_name: str,
    feedback_key: str,
) -> SimpleAsyncEvaluator:
    if code_extractor is not None and code_extraction_strategy != "none":
        raise ValueError(
            "`code_extractor` and `code_extraction_strategy` cannot both be provided"
        )

    if code_extraction_strategy == "llm":
        if model is None and client is None:
            raise ValueError("You must provide either a `model` string or a `client`")
        if client is None:
            client = init_chat_model(model)  # type: ignore

    async def _wrapped_evaluator(
        *,
        inputs: Optional[Union[str, dict]] = None,
        outputs: Union[str, dict],
        reference_outputs: Optional[Union[str, dict]] = None,
        **kwargs,
    ):
        async def _ascore_wrapper(*, outputs: Union[str, dict], **kwargs):
            if code_extractor is None:
                normalized_outputs = _normalize_final_app_outputs_as_string(outputs)
                if code_extraction_strategy == "llm":
                    model_with_tools = client.bind_tools([ExtractCode, NoCode])  # type: ignore
                    res = await model_with_tools.ainvoke(
                        [
                            {"role": "system", "content": LLM_EXTRACTION_SYSTEM_PROMPT},
                            {
                                "role": "user",
                                "content": LLM_EXTRACTION_USER_PROMPT.format(
                                    outputs=normalized_outputs
                                ),
                            },
                        ],
                        {"run_name": "extract_code"},
                    )
                    if res.tool_calls[0]["name"] == "ExtractCode":  # type: ignore
                        normalized_outputs = res.tool_calls[0]["args"]["code"]  # type: ignore
                    else:
                        return (False, None, {"code_extraction_failed": True})
                elif code_extraction_strategy == "markdown_code_blocks":
                    normalized_outputs = _extract_code_from_markdown_code_blocks(  # type: ignore
                        normalized_outputs
                    )
                    if normalized_outputs is None:
                        return (False, None, {"code_extraction_failed": True})
                else:
                    # Nothing to do to extract code
                    pass
            else:
                normalized_outputs = code_extractor(outputs)  # type: ignore
                if hasattr(normalized_outputs, "__await__"):
                    normalized_outputs = await normalized_outputs
            score_result = scorer(
                outputs=normalized_outputs,
                **kwargs,
            )
            if hasattr(score_result, "__await__"):
                return await score_result
            return score_result

        return await _arun_evaluator(
            run_name=run_name,
            scorer=_ascore_wrapper,
            feedback_key=feedback_key,
            inputs=inputs,
            outputs=outputs,
            reference_outputs=reference_outputs,
            **kwargs,
        )

    return _wrapped_evaluator  # type: ignore



import json
import pytest

from openevals.llm import create_async_llm_as_judge

from langchain_core.messages import HumanMessage
from openai import AsyncOpenAI
from langchain_openai import ChatOpenAI
from langsmith import Client
from langchain import hub as prompts


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_prompt_hub_works():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt=prompts.pull("langchain-ai/test-equality"),
        judge=client,
        model="gpt-4o-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] is not None
    assert eval_result["comment"] is not None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_prompt_hub_works_one_message():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt=prompts.pull("langchain-ai/equality-1-message"),
        judge=client,
        model="gpt-4o-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] is not None
    assert eval_result["comment"] is not None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
        model="gpt-4o-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] is not None
    assert eval_result["comment"] is not None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai_no_reasoning():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
        model="gpt-4o-mini",
        use_reasoning=False,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] is not None
    assert eval_result["comment"] is None


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai_not_equal():
    inputs = {"a": 1, "b": 3}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
        model="gpt-4o-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai_not_equal_continuous():
    inputs = {"a": 1, "b": 3}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="How equal are these 2? {inputs} {outputs}",
        judge=client,
        model="gpt-4o-mini",
        continuous=True,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] > 0
    assert eval_result["score"] < 1


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai_not_equal_binary_fail():
    inputs = {"a": 1, "b": 3}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
        model="o3-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_openai_not_equal_binary_pass():
    inputs = {"a": 1, "b": 3}
    outputs = {"a": 1, "b": 2}
    client = AsyncOpenAI()
    llm_as_judge = create_async_llm_as_judge(
        prompt="How equal are these 2? {inputs} {outputs}",
        judge=client,
        model="o3-mini",
        continuous=True,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"] > 0


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_langchain():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    client = ChatOpenAI(model="gpt-4o-mini")
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_llm_as_judge_langchain_messages():
    inputs = [HumanMessage(content=json.dumps({"a": 1, "b": 2}))]
    outputs = [HumanMessage(content=json.dumps({"a": 1, "b": 3}))]
    client = ChatOpenAI(model="gpt-4o-mini")
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_llm_as_judge_langchain_messages_dict():
    inputs = {"messages": [HumanMessage(content=json.dumps({"a": 1, "b": 2}))]}
    outputs = {"messages": [HumanMessage(content=json.dumps({"a": 1, "b": 3}))]}
    client = ChatOpenAI(model="gpt-4o-mini")
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        judge=client,
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_init_chat_model():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two equal? {inputs} {outputs}",
        model="openai:o3-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_few_shot_examples():
    inputs = {"a": 1, "b": 2}
    outputs = {"a": 1, "b": 2}
    llm_as_judge = create_async_llm_as_judge(
        prompt="Are these two foo? {inputs} {outputs}",
        few_shot_examples=[
            {"inputs": {"a": 1, "b": 2}, "outputs": {"a": 1, "b": 2}, "score": 0.0},
            {"inputs": {"a": 1, "b": 3}, "outputs": {"a": 1, "b": 2}, "score": 1.0},
        ],
        model="openai:o3-mini",
    )
    eval_result = await llm_as_judge(inputs=inputs, outputs=outputs)
    assert not eval_result["score"]


@pytest.mark.langsmith
@pytest.mark.asyncio
async def test_async_llm_as_judge_with_evaluate():
    client = Client()
    evaluator = create_async_llm_as_judge(
        prompt="Are these two foo? {inputs} {outputs}",
        few_shot_examples=[
            {"inputs": {"a": 1, "b": 2}, "outputs": {"a": 1, "b": 2}, "score": 0.0},
            {"inputs": {"a": 1, "b": 3}, "outputs": {"a": 1, "b": 2}, "score": 1.0},
        ],
        model="openai:o3-mini",
    )

    async def target(x):
        return x

    res = await client.aevaluate(target, data="exact match", evaluators=[evaluator])
    async for r in res:
        assert r["evaluation_results"]["results"][0].score is not None



import json
import subprocess
import tempfile
import os

from typing import Optional, Callable, Any, Union, Literal, Awaitable, Tuple

from openevals.code.llm import (
    _create_base_code_evaluator,
    _create_async_base_code_evaluator,
)
from openevals.types import SimpleEvaluator, SimpleAsyncEvaluator

from langchain_core.language_models.chat_models import BaseChatModel


def _parse_pyright_output(stdout: bytes) -> Tuple[bool, str]:
    try:
        # Parse the JSON output
        output = json.loads(stdout)

        errors = []
        for error in output.get("generalDiagnostics", []):
            if (
                error.get("severity", None) == "error"
                and error.get("rule", None) != "reportMissingImports"
            ):
                del error["file"]
                errors.append(error)
        score = len(errors) == 0
        return (score, json.dumps(errors))
    except json.JSONDecodeError:
        return (False, f"Failed to parse Pyright output: {stdout.decode()}")


def _analyze_with_pyright(
    *,
    output: str,
    pyright_cli_args: list[str],
):
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".py", mode="w", delete=False) as temp:
        temp.write(output)
        temp_path = temp.name

    try:
        result = subprocess.run(
            [
                "pyright",
                "--outputjson",
                "--level",
                "error",  # Only report errors, not warnings
                *(pyright_cli_args or []),
                temp_path,
            ],
            capture_output=True,
        )

        return _parse_pyright_output(result.stdout)  # type: ignore

    finally:
        # Clean up
        os.unlink(temp_path)


async def _analyze_with_pyright_async(
    *,
    output: str,
    pyright_cli_args: list[str],
):
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".py", mode="w", delete=False) as temp:
        temp.write(output)
        temp_path = temp.name

    try:
        # Use asyncio.create_subprocess_exec for async subprocess execution
        import asyncio

        process = await asyncio.create_subprocess_exec(
            "pyright",
            "--outputjson",
            "--level",
            "error",  # Only report errors, not warnings
            *(pyright_cli_args or []),
            temp_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        stdout, _ = await process.communicate()
        return _parse_pyright_output(stdout)  # type: ignore

    finally:
        # Clean up
        os.unlink(temp_path)


def create_pyright_evaluator(
    *,
    pyright_cli_args: list[str] = [],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], str]] = None,
    client: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
) -> SimpleEvaluator:
    """Creates an evaluator that checks Python code using Pyright.

    This evaluator runs Pyright static type checking on Python code and returns whether
    the code passes type checking along with any error details.

    Args:
        pyright_cli_args: Command-line arguments to pass to Pyright.
        code_extraction_strategy: Strategy for extracting code from the output.
            - "none": Use the entire output as code.
            - "llm": Use an LLM to extract code from the output.
            - "markdown_code_blocks": Extract code from markdown code blocks.
        code_extractor: Custom function to extract code from the output.
            Can be synchronous or asynchronous.
        client: LLM client to use for code extraction if strategy is "llm".
        model: Model name to use for code extraction if strategy is "llm".

    Returns:
        An evaluator function.

    Raises:
        ValueError: If client or model is provided with a code_extraction_strategy
                   other than "llm".
    """
    if code_extraction_strategy != "llm" and (client or model):
        raise ValueError(
            "client and model may only be passed if code_extraction_strategy is 'llm'"
        )

    def _scorer(
        *,
        outputs: str,
        **kwargs,
    ):
        return _analyze_with_pyright(
            output=outputs,
            pyright_cli_args=pyright_cli_args,
        )

    return _create_base_code_evaluator(
        model=model,
        client=client,
        run_name="pyright_evaluator",
        feedback_key="pyright_succeeded",
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )


def create_async_pyright_evaluator(
    *,
    pyright_cli_args: list[str] = [],
    code_extraction_strategy: Literal["none", "llm", "markdown_code_blocks"] = "none",
    code_extractor: Optional[Callable[[Any], Union[str, Awaitable[str]]]] = None,
    client: Optional[BaseChatModel] = None,
    model: Optional[str] = None,
) -> SimpleAsyncEvaluator:
    """Creates an asynchronous evaluator that checks Python code using Pyright.

    This evaluator runs Pyright static type checking on Python code and returns whether
    the code passes type checking along with any error details.

    Args:
        pyright_cli_args: Command-line arguments to pass to Pyright.
        code_extraction_strategy: Strategy for extracting code from the output.
            - "none": Use the entire output as code.
            - "llm": Use an LLM to extract code from the output.
            - "markdown_code_blocks": Extract code from markdown code blocks.
        code_extractor: Custom function to extract code from the output.
            Can be synchronous or asynchronous.
        client: LLM client to use for code extraction if strategy is "llm".
        model: Model name to use for code extraction if strategy is "llm".

    Returns:
        An asynchronous evaluator function.

    Raises:
        ValueError: If client or model is provided with a code_extraction_strategy
                   other than "llm".
    """
    if code_extraction_strategy != "llm" and (client or model):
        raise ValueError(
            "client and model may only be passed if code_extraction_strategy is 'llm'"
        )

    async def _scorer(
        *,
        outputs: str,
        **kwargs,
    ):
        return await _analyze_with_pyright_async(
            output=outputs,
            pyright_cli_args=pyright_cli_args,
        )

    return _create_async_base_code_evaluator(
        model=model,
        client=client,
        run_name="pyright_check",
        feedback_key="pyright_succeeded",
        scorer=_scorer,
        code_extraction_strategy=code_extraction_strategy,
        code_extractor=code_extractor,
    )
